# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial
from .framework import api_select


def test_direct_url_is_public_url(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=1, params=(('s1', 'Hi', 'desc'), ('c1', ('true', 'maybe', 'false'), 'desc')))
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=2, serial=True)

        with serial(api, timeout=40, job_name_prefix=api.job_name_prefix, report_interval=1, direct_url=api.public_uri) as ctrl:
            ctrl.invoke('j1', s1='HELLO', c1=True)
            ctrl.invoke('j2')

    sout, _ = capsys.readouterr()
    assert '//job' not in sout
    assert '/job/' in sout


def test_direct_url_trailing_slash(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=1, params=(('s1', 'Hi', 'desc'),))
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=2, serial=True)

        with serial(api, timeout=40, job_name_prefix=api.job_name_prefix, report_interval=1, direct_url=api.public_uri + '/') as ctrl:
            ctrl.invoke('j1', s1='HELLO')
            ctrl.invoke('j2')

    sout, _ = capsys.readouterr()
    assert '//job' not in sout
    assert '/job/' in sout



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os

import pytest
from pytest import raises

from jenkinsflow.flow import parallel, serial, FailedChildJobException, FailedChildJobsException

from .framework import api_select
from .framework.abort_job import abort
from .cfg import ApiType

here = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.not_apis(ApiType.SCRIPT)
def test_abort_retry_serial_toplevel(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j12_abort', max_fails=0, expect_invocations=1, expect_order=2, exec_time=20, serial=True, final_result='ABORTED')
        api.job('j13', max_fails=0, expect_invocations=0, expect_order=None, serial=True)

        abort(api, 'j12_abort', 10)

        with raises(FailedChildJobException):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
                ctrl1.invoke('j11')
                ctrl1.invoke('j12_abort')
                ctrl1.invoke('j13')


@pytest.mark.not_apis(ApiType.SCRIPT)
def test_abort_retry_parallel_toplevel(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('j12_abort', max_fails=0, expect_invocations=1, expect_order=None, exec_time=20, final_result='ABORTED')
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=None)

        abort(api, 'j12_abort', 10)

        with raises(FailedChildJobsException):
            with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
                ctrl1.invoke('j11')
                ctrl1.invoke('j12_abort')
                ctrl1.invoke('j13')


@pytest.mark.not_apis(ApiType.SCRIPT)
def test_abort_retry_serial_parallel_nested(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2, exec_time=20)
        api.job('j22_abort', max_fails=0, expect_invocations=1, expect_order=2, exec_time=20, final_result='ABORTED')
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=2, exec_time=5)
        api.job('j24', max_fails=0, expect_invocations=1, expect_order=2)
        api.job('j12', max_fails=0, expect_invocations=0, expect_order=None, serial=True)

        abort(api, 'j22_abort', 10)

        with raises(FailedChildJobException):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as sctrl1:
                sctrl1.invoke('j11')
                with sctrl1.parallel() as pctrl1:
                    pctrl1.invoke('j21')
                    pctrl1.invoke('j22_abort')
                    pctrl1.invoke('j23')
                    pctrl1.invoke('j24')
                sctrl1.invoke('j12')


@pytest.mark.not_apis(ApiType.SCRIPT)
def test_abort_retry_parallel_serial_nested(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('j22_abort', max_fails=0, expect_invocations=1, expect_order=None, exec_time=20, final_result='ABORTED')
        api.job('j23', max_fails=0, expect_invocations=0, expect_order=None)
        api.job('j24', max_fails=0, expect_invocations=0, expect_order=None)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=1, serial=True)

        abort(api, 'j22_abort', 10)

        with raises(FailedChildJobsException):
            with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as sctrl1:
                sctrl1.invoke('j11')
                with sctrl1.serial() as pctrl1:
                    pctrl1.invoke('j21')
                    pctrl1.invoke('j22_abort')
                    pctrl1.invoke('j23')
                    pctrl1.invoke('j24')
                sctrl1.invoke('j12')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pytest import raises

from jenkinsflow.flow import serial, FailedChildJobException
from .framework import api_select


def test_multi_level_errors(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('wait2', max_fails=0, expect_invocations=1, expect_order=1, exec_time=2)
        api.job('wait5', max_fails=0, expect_invocations=1, expect_order=2, exec_time=5)
        api.job('quick_fail', max_fails=1, expect_invocations=1, expect_order=2, params=(('s1', 'WORLD', 'desc'), ('c1', ('why', 'not'), 'desc')))
        api.job('not_invoked', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(FailedChildJobException):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
                ctrl1.invoke('wait2')

                with ctrl1.parallel(timeout=20, report_interval=3) as ctrl2:
                    ctrl2.invoke('wait5')
                    ctrl2.invoke('quick_fail', password='Y', s1='WORLD', c1='why')

                # Never invoked because of failure in preceding 'parallel'
                ctrl1.invoke('not_invoked')



import sys

from .api_base import AuthError


major_version = sys.version_info.major


def encode(text, encoding):
    if major_version >= 3:
        # TODO: Why is this necessary???
        # On python 2 we need to convert from bytes to unicode, but on python 3 we need the reverse!
        return text.encode("utf-8")
    try:
        # TODO: 'unicode' call was needed when implemented, but now seems to be (randomly) raisig TypeError
        # and not actually needed???
        return unicode(text, encoding, errors='ignore')  # pylint: disable=undefined-variable
    except TypeError:
        pass
    return text


class ResourceNotFound(Exception):
    pass


if major_version < 3:
    # ConnectionError is builtin in Python 3
    class ConnectionError(Exception):
        pass


class RequestsRestApi(object):
    def __init__(self, direct_uri, username, password):
        super(RequestsRestApi, self).__init__()
        import requests
        self.session = requests.Session()
        if username or password:
            self.session.auth = requests.auth.HTTPBasicAuth(username, encode(password, 'latin-1'))
        self.direct_uri = direct_uri

    @staticmethod
    def _check_response(response, good_responses=(200,)):
        if response.status_code in good_responses:
            return response

        try:
            response.raise_for_status()
        except Exception as ex:
            if response.status_code == 404:
                raise ResourceNotFound(ex)
            if response.status_code in (401, 403):
                raise AuthError(ex)
            raise

        # TODO: This is dubious, maybe we shoud raise here instead.
        return response

    def _get(self, url, params):
        import requests
        try:
            return self._check_response(self.session.get(self.direct_uri + url, params=params))
        except requests.ConnectionError as ex:
            raise ConnectionError(ex)

    def get_content(self, url, **params):
        return self._get(url, params=params).content

    def get_json(self, url="", **params):
        return self._get(url + "/api/json", params=params).json()

    def post(self, url, payload=None, headers=None, allow_redirects=False, **params):
        response = self.session.post(self.direct_uri + url, headers=headers, data=payload, allow_redirects=allow_redirects, params=params)
        return self._check_response(response, (200, 201))

    def headers(self, allow_redirects=True):
        return self._check_response(self.session.head(self.direct_uri, allow_redirects=allow_redirects)).headers


def _check_restkit_response(func):
    def deco(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except self.restkit.Unauthorized as ex:
            raise AuthError(ex.response.status + " user: '" + self.username + "' for url: " + ex.response.final_url)
        except self.restkit.errors.ResourceNotFound as ex:
            raise ResourceNotFound(str(ex))

    return deco


class RestkitRestApi(object):
    def __init__(self, direct_uri, username, password, **kwargs):
        super(RestkitRestApi, self).__init__()

        import restkit
        self.restkit = restkit
        import json
        self.json = json

        if username or password:
            filters = kwargs.get('filters', [])
            filters.append(restkit.BasicAuth(username, password))
            kwargs['filters'] = filters

        self.username = username
        self.session = restkit.Resource(uri=direct_uri, **kwargs)

    @_check_restkit_response
    def get_content(self, url, **params):
        return self.session.get(url, **params).body_string()

    @_check_restkit_response
    def get_json(self, path="", headers=None, params_dict=None, **params):
        response = self.session.get(path=path + "/api/json", headers=headers, params_dict=params_dict, **params)
        return self.json.loads(response.body_string())

    @_check_restkit_response
    def post(self, url, payload=None, headers=None, allow_redirects=False, **params):  # pylint: disable=unused-argument
        return self.session.post(url, headers=headers, payload=payload, **params)

    @_check_restkit_response
    def headers(self, allow_redirects=True):  # pylint: disable=unused-argument
        return self.session.head().headers



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import time

import pytest

from jenkinsflow.speed import Speed

from .cfg import ApiType


@pytest.mark.not_apis(ApiType.MOCK)
def test_hyperspeed_speedup():
    hs = Speed()
    assert hs.speedup == 1
    

@pytest.mark.not_apis(ApiType.MOCK)
def test_hyperspeed_real_time():
    hs = Speed()
    assert hs.time() <= time.time()


@pytest.mark.not_apis(ApiType.MOCK)
def test_hyperspeed_real_sleep():
    hs = Speed()
    before = time.time()
    hs.sleep(1)
    after = time.time()
    assert after - before >= 1



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial, parallel
from .framework import api_select


def test_retry_unchecked_alone_serial_toplevel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11_unchecked', max_fails=1, expect_invocations=1, expect_order=1, exec_time=20, invocation_delay=0, unknown_result=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            ctrl1.invoke_unchecked('j11_unchecked')


def test_retry_unchecked_long_running_serial_toplevel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11_unchecked', max_fails=1, expect_invocations=1, expect_order=1, exec_time=20, unknown_result=True)
        api.job('j12_fail', max_fails=1, expect_invocations=2, expect_order=2)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=3, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            ctrl1.invoke_unchecked('j11_unchecked')
            ctrl1.invoke('j12_fail')
            ctrl1.invoke('j13')


def test_retry_unchecked_long_running_parallel_toplevel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11_unchecked', max_fails=1, expect_invocations=1, expect_order=1, exec_time=20, unknown_result=True)
        api.job('j12_fail', max_fails=1, expect_invocations=2, expect_order=1)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=1)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            ctrl1.invoke_unchecked('j11_unchecked')
            ctrl1.invoke('j12_fail')
            ctrl1.invoke('j13')


def test_retry_unchecked_quick_serial_toplevel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11_unchecked', max_fails=1, expect_invocations=2, expect_order=1)
        api.job('j12_fail', max_fails=1, expect_invocations=2, expect_order=2, exec_time=1)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=3, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            ctrl1.invoke_unchecked('j11_unchecked')
            ctrl1.invoke('j12_fail')
            ctrl1.invoke('j13')


def test_retry_unchecked_quick_parallel_toplevel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11_unchecked', max_fails=1, expect_invocations=2, expect_order=1)
        api.job('j12_fail', max_fails=1, expect_invocations=2, expect_order=1, exec_time=1)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=1)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            ctrl1.invoke_unchecked('j11_unchecked')
            ctrl1.invoke('j12_fail')
            ctrl1.invoke('j13')


def test_retry_unchecked_quick_serial_outer_level(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11_unchecked', max_fails=2, expect_invocations=3, expect_order=1)
        api.job('j12_fail', max_fails=2, expect_invocations=3, expect_order=2, exec_time=1)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=3, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            with ctrl1.serial(max_tries=2) as ctrl2:
                ctrl2.invoke_unchecked('j11_unchecked')
                ctrl2.invoke('j12_fail')
                ctrl2.invoke('j13')


def test_retry_unchecked_quick_parallel_outer_level(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j11_unchecked', max_fails=2, expect_invocations=3, expect_order=1)
        api.job('j12_fail', max_fails=2, expect_invocations=3, expect_order=1, exec_time=1)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=1)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            with ctrl1.parallel(max_tries=2) as ctrl2:
                ctrl2.invoke_unchecked('j11_unchecked')
                ctrl2.invoke('j12_fail')
                ctrl2.invoke('j13')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial
from .framework import api_select


def test_no_args(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('job-1', max_fails=0, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke('job-1')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pytest import raises

from jenkinsflow.flow import parallel, serial, FlowTimeoutException
from .framework import api_select


def test_timeout_top_level_serial(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.job('quick', max_fails=0, expect_invocations=1, expect_order=None, params=(('s1', '', 'desc'), ('c1', 'false', 'desc')))
        api.job('wait20', max_fails=0, expect_invocations=1, expect_order=None, exec_time=20, unknown_result=True, serial=True)

        with raises(FlowTimeoutException):
            with serial(api, timeout=8, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl:
                ctrl.invoke('quick', s1='', c1=False)
                ctrl.invoke('wait20')


def test_timeout_top_level_parallel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.job('quick', max_fails=0, expect_invocations=1, expect_order=None, params=(('s1', '', 'desc'), ('c1', 'false', 'desc')))
        api.job('wait20', max_fails=0, expect_invocations=1, expect_order=None, exec_time=20, unknown_result=True)

        with raises(FlowTimeoutException):
            with parallel(api, timeout=8, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl:
                ctrl.invoke('quick', s1='', c1=False)
                ctrl.invoke('wait20')


def test_timeout_inner_level_serial(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.job('quick11', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('quick21', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('wait20', max_fails=0, expect_invocations=1, expect_order=None, exec_time=20, unknown_result=True, serial=True)

        with raises(FlowTimeoutException) as exinfo:
            with parallel(api, timeout=3000, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
                ctrl1.invoke('quick11')
                with ctrl1.serial(timeout=8) as ctrl2:
                    ctrl2.invoke('quick21')
                    ctrl2.invoke('wait20')

        assert "Timeout after:" in str(exinfo.value)
        assert ", in flow ['jenkinsflow_test__timeout_inner_level_serial__quick21', 'jenkinsflow_test__timeout_inner_level_serial__wait20']. Unfinished jobs:['jenkinsflow_test__timeout_inner_level_serial__wait20']" in str(exinfo.value)



def test_timeout_inner_level_parallel(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.job('quick11', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('quick21', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('wait20', max_fails=0, expect_invocations=1, expect_order=None, exec_time=30, unknown_result=True)

        with raises(FlowTimeoutException) as exinfo:
            with serial(api, timeout=3000, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
                ctrl1.invoke('quick11')
                with ctrl1.parallel(timeout=8) as ctrl2:
                    ctrl2.invoke('quick21')
                    ctrl2.invoke('wait20')

        assert "Timeout after:" in str(exinfo.value)
        assert ", in flow ('jenkinsflow_test__timeout_inner_level_parallel__quick21', 'jenkinsflow_test__timeout_inner_level_parallel__wait20'). Unfinished jobs:['jenkinsflow_test__timeout_inner_level_parallel__wait20']" in str(exinfo.value)


def test_timeout_multi_level_mix(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.job('quick11', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('quick21', max_fails=0, expect_invocations=1, expect_order=None)
        api.job('wait20_22', max_fails=0, expect_invocations=1, expect_order=None, exec_time=30, unknown_result=True)
        api.job('wait20_31', max_fails=0, expect_invocations=1, expect_order=None, exec_time=30, unknown_result=True)

        with raises(FlowTimeoutException) as exinfo:
            with serial(api, timeout=3000, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl1:
                ctrl1.invoke('quick11')
                with ctrl1.parallel(timeout=8) as ctrl2:
                    ctrl2.invoke('quick21')
                    ctrl2.invoke('wait20_22')
                    with ctrl2.parallel() as ctrl3:
                        ctrl3.invoke('wait20_31')

        assert "Timeout after:" in str(exinfo.value)
        assert ", in flow ('jenkinsflow_test__timeout_multi_level_mix__quick21', 'jenkinsflow_test__timeout_multi_level_mix__wait20_22', ('jenkinsflow_test__timeout_multi_level_mix__wait20_31',)). Unfinished jobs:['jenkinsflow_test__timeout_multi_level_mix__wait20_22', ('jenkinsflow_test__timeout_multi_level_mix__wait20_31',)]" in str(exinfo.value)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function

import click

from ..jenkins_cli_protocol import CliProtocol
from ..utils import set_build_result as usbr


@click.command()
@click.option('--result', help="The result to set. Should probably be 'unstable'", default='unstable')
@click.option('--username', help="Name of jenkins user with access to the job")
@click.option('--password', help="Password of jenkins user with access to the job.")
@click.option('--java', help="Alternative 'java' executable", default='java')
@click.option('--direct-url', help="Jenkins URL. Default is JENKINS_URL/HUDSON_URL env var value. Use this argument if JENKINS_URL is a proxy [default: None]")
@click.option(
    '--protocol', default='remoting',
    help="The cli protocol to use. See https://jenkins.io/doc/book/managing/cli/. Only 'remoting' or unspecified is supported. [default: 'remoting'.]" \
    " Specify '' to not supply this option to the cli (for older Jenkins")
def set_build_result(result, username, password, direct_url, java, protocol):
    """Change the result of a Jenkins job.

    DEPRECATED - You should use shell step exit code to determine result.

    Note: set_build_result can only be done from within the job, not after the job has finished.
    Note: Only available if URL is set in `Jenkins <http://jenkins-ci.org/>`_ system configuration.

    This command uses the Jenkins `cli` to change the result. It requires a java executable to run the Jenkins `cli`.
    Please note that in some versions of jenkins the cli is broken, it has no manifest file!
    """
    # %(file)s [--result <result>] [--java <java>] [--direct-url <direct_url>] [(--username <user_name> --password <password>)]

    usbr.set_build_result(result, username, password, direct_url, java, CliProtocol[protocol] if protocol else None)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function

import os

import pytest

from jenkinsflow.flow import serial

from .framework import api_select
from .cfg import ApiType


@pytest.mark.apis(ApiType.SCRIPT)
def test_script_api_env_unchanged(api_type):
    """Ensure that os.environ is unchanged after run of slow using script_api"""

    env_before = os.environ.copy()

    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        job_name = 'job-1'
        api.job(job_name, max_fails=0, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl1:
            ctrl1.invoke(job_name)

        # Do a few more things to 
        job = api.get_job(api.job_name_prefix + job_name)
        _, _, build_num = job.job_status()

    assert os.environ == env_before



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os
from pytest import raises

from jenkinsflow.flow import parallel, serial, BuildResult, FailedChildJobException

from demo_security import username, password
from .framework import api_select
from .framework.utils import pre_existing_fake_cli


def test_propagate_unstable_serial_single_unstable(api_type, fake_java):
    with api_select.api(__file__, api_type, login=True) as api:
        pre_existing_fake_cli(api_type)
        api.flow_job()
        api.job('j11_unstable', max_fails=0, expect_invocations=1, expect_order=1, final_result='unstable')

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11_unstable')

        assert ctrl1.result == BuildResult.UNSTABLE


def test_propagate_unstable_serial_single_unstable_user_pass(api_type, fake_java):
    with api_select.api(__file__, api_type) as api:
        pre_existing_fake_cli(api_type)
        api.flow_job()
        api.job('j11_unstable', max_fails=0, expect_invocations=1, expect_order=1, final_result='unstable')

        with serial(api, timeout=70, username=username, password=password, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11_unstable')

        assert ctrl1.result == BuildResult.UNSTABLE


def test_propagate_unstable_parallel_single_unstable(api_type, fake_java):
    with api_select.api(__file__, api_type, login=True) as api:
        pre_existing_fake_cli(api_type)
        api.flow_job()
        api.job('j11_unstable', max_fails=0, expect_invocations=1, expect_order=1, final_result='unstable')

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11_unstable')

        assert ctrl1.result == BuildResult.UNSTABLE


def test_propagate_unstable_serial_toplevel(api_type, fake_java):
    with api_select.api(__file__, api_type, login=True) as api:
        pre_existing_fake_cli(api_type)
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j12_unstable', max_fails=0, expect_invocations=1, expect_order=2, final_result='unstable', serial=True)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=3, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            ctrl1.invoke('j12_unstable')
            ctrl1.invoke('j13')

        assert ctrl1.result == BuildResult.UNSTABLE


def test_propagate_unstable_parallel_toplevel(api_type, fake_java):
    with api_select.api(__file__, api_type, login=True) as api:
        pre_existing_fake_cli(api_type)
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j12_unstable', max_fails=0, expect_invocations=1, expect_order=1, final_result='unstable')
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=1)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            ctrl1.invoke('j12_unstable')
            ctrl1.invoke('j13')

        assert ctrl1.result == BuildResult.UNSTABLE


def test_propagate_unstable_serial_inner(api_type, fake_java):
    with api_select.api(__file__, api_type, login=True) as api:
        pre_existing_fake_cli(api_type)
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2, serial=True)
        api.job('j22_unstable', max_fails=0, expect_invocations=1, expect_order=3, final_result='unstable', serial=True)
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=4, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.serial() as ctrl2:
                ctrl2.invoke('j21')
                ctrl2.invoke('j22_unstable')
                ctrl2.invoke('j23')

        assert ctrl2.result == BuildResult.UNSTABLE
        assert ctrl1.result == BuildResult.UNSTABLE


def test_propagate_unstable_parallel_inner(api_type, fake_java):
    with api_select.api(__file__, api_type, login=True) as api:
        pre_existing_fake_cli(api_type)
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2, serial=True)
        api.job('j22_unstable', max_fails=0, expect_invocations=1, expect_order=2, final_result='unstable')
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=2)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.parallel() as ctrl2:
                ctrl2.invoke('j21')
                ctrl2.invoke('j22_unstable')
                ctrl2.invoke('j23')

        assert ctrl2.result == BuildResult.UNSTABLE
        assert ctrl1.result == BuildResult.UNSTABLE


def test_propagate_unstable_serial_inner_fail_after(api_type, fake_java):
    with api_select.api(__file__, api_type, login=True) as api:
        pre_existing_fake_cli(api_type)
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2)
        api.job('j22_unstable', max_fails=0, expect_invocations=1, expect_order=3, final_result='unstable')
        api.job('j23_fail', max_fails=1, expect_invocations=1, expect_order=4)

        with raises(FailedChildJobException):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
                ctrl1.invoke('j11')
                with ctrl1.serial() as ctrl2:
                    ctrl2.invoke('j21')
                    ctrl2.invoke('j22_unstable')
                    ctrl2.invoke('j23_fail')


def test_propagate_unstable_parallel_inner_fail_before(api_type, fake_java):
    with api_select.api(__file__, api_type, login=True) as api:
        pre_existing_fake_cli(api_type)
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21_fail', max_fails=1, expect_invocations=1, expect_order=2, serial=True)
        api.job('j22_unstable', max_fails=0, expect_invocations=1, expect_order=2, final_result='unstable')
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=2)

        with raises(FailedChildJobException):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
                ctrl1.invoke('j11')
                with ctrl1.parallel() as ctrl2:
                    ctrl2.invoke('j21_fail')
                    ctrl2.invoke('j22_unstable')
                    ctrl2.invoke('j23')



# -*- coding: utf-8 -*-

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import pytest

from jenkinsflow import flow
from jenkinsflow.api_base import AuthError

from .framework import api_select
from .cfg import ApiType


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_auth_error(api_type):
    with api_select.api(__file__, api_type, username="noaccess", password='maybegranted') as api:

        with pytest.raises(AuthError) as exinfo:
            with flow.serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
                ctrl1.invoke('j11')

        assert ("401 Client Error: Invalid password/token for user: noaccess for url: http://" in str(exinfo.value) or
                "401 Unauthorized user: 'noaccess' for url: http://" in str(exinfo.value))



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pytest import raises

from jenkinsflow.flow import parallel, serial, FailedChildJobException, FailedChildJobsException
from .framework import api_select, utils


def test_retry_serial_toplevel(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=2, expect_order=1)
        api.job('j12_fail', max_fails=1, expect_invocations=2, expect_order=2, serial=True)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=3, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            ctrl1.invoke('j11')
            ctrl1.invoke('j12_fail')
            ctrl1.invoke('j13')


def test_retry_serial_toplevel_fail(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=2, expect_order=1)
        api.job('j12_fail', max_fails=2, expect_invocations=2, expect_order=2, serial=True)
        api.job('j13', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(FailedChildJobException):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
                ctrl1.invoke('j11')
                ctrl1.invoke('j12_fail')
                ctrl1.invoke('j13')


def test_retry_parallel_toplevel(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j12_fail', max_fails=1, expect_invocations=2, expect_order=1)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=1)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            ctrl1.invoke('j11')
            ctrl1.invoke('j12_fail')
            ctrl1.invoke('j13')


def test_retry_parallel_toplevel_fail(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j12_fail', max_fails=2, expect_invocations=2, expect_order=1)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=1)

        with raises(FailedChildJobsException):
            with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
                ctrl1.invoke('j11')
                ctrl1.invoke('j12_fail')
                ctrl1.invoke('j13')


def test_retry_serial_inner(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=2, expect_order=2, serial=True)
        api.job('j22_fail', max_fails=1, expect_invocations=2, expect_order=3, serial=True)
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=4, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.serial(max_tries=2) as ctrl2:
                ctrl2.invoke('j21')
                ctrl2.invoke('j22_fail')
                ctrl2.invoke('j23')


def test_retry_parallel_inner(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2)
        api.job('j22_fail', max_fails=1, expect_invocations=2, expect_order=2)
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=2)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.parallel(max_tries=2) as ctrl2:
                ctrl2.invoke('j21')
                ctrl2.invoke('j22_fail')
                ctrl2.invoke('j23')


def test_retry_serial_through_parent_serial_level(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=2, expect_order=2, serial=True)
        api.job('j31_fail', max_fails=2, expect_invocations=3, expect_order=3, serial=True)
        api.job('j32', max_fails=0, expect_invocations=1, expect_order=4, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.serial(timeout=70, max_tries=2) as ctrl2:
                ctrl2.invoke('j21')
                with ctrl2.serial(timeout=70, max_tries=2) as ctrl3:
                    ctrl3.invoke('j31_fail')
                    ctrl3.invoke('j32')


def test_retry_serial_through_parent_serial_level_fail(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j31_fail', max_fails=4, expect_invocations=4, expect_order=1)
        api.job('j32', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(FailedChildJobException):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
                with ctrl1.serial(timeout=70, max_tries=2) as ctrl2:
                    with ctrl2.serial(timeout=70, max_tries=2) as ctrl3:
                        ctrl3.invoke('j31_fail')
                        ctrl3.invoke('j32')


def test_retry_serial_through_parent_parallel_level(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2, serial=True)
        api.job('j31_fail', max_fails=2, expect_invocations=3, expect_order=2)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.parallel(timeout=70, max_tries=2) as ctrl2:
                ctrl2.invoke('j21')
                with ctrl2.serial(timeout=70, max_tries=2) as ctrl3:
                    ctrl3.invoke('j31_fail')


def test_retry_serial_through_parent_parallel_level_fail(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j31_fail', max_fails=4, expect_invocations=4, expect_order=1)

        with raises(FailedChildJobException):
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
                with ctrl1.parallel(timeout=70, max_tries=2) as ctrl2:
                    with ctrl2.serial(timeout=70, max_tries=2) as ctrl3:
                        ctrl3.invoke('j31_fail')


def test_retry_parallel_through_parent_serial_level(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=2, expect_order=2)
        api.job('j31_fail', max_fails=2, expect_invocations=3, expect_order=3)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.serial(timeout=70, max_tries=2) as ctrl2:
                ctrl2.invoke('j21')
                with ctrl2.parallel(timeout=70, max_tries=2) as ctrl3:
                    ctrl3.invoke('j31_fail')


def test_retry_parallel_through_parent_parallel_level(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2, serial=True)
        api.job('j22_fail', max_fails=1, expect_invocations=2, expect_order=2)
        api.job('j31_fail', max_fails=2, expect_invocations=3, expect_order=2)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.parallel(timeout=70, max_tries=2) as ctrl2:
                ctrl2.invoke('j21')
                ctrl2.invoke('j22_fail')
                with ctrl2.parallel(timeout=70, max_tries=2) as ctrl3:
                    ctrl3.invoke('j31_fail')


def test_retry_parallel_through_outer_level(api_type, capsys):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2, serial=True)
        api.job('j31', max_fails=0, expect_invocations=1, expect_order=3, serial=True)
        api.job('j41_fail', max_fails=2, expect_invocations=3, expect_order=3)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.serial(timeout=70, max_tries=2) as ctrl2:
                ctrl2.invoke('j21')
                with ctrl2.parallel(timeout=70) as ctrl3:
                    ctrl3.invoke('j31')
                    with ctrl3.parallel(timeout=70, max_tries=2) as ctrl4:
                        ctrl4.invoke('j41_fail')

        sout, _ = capsys.readouterr()
        sout = utils.replace_host_port(api_type, sout)
        assert "RETRY: job: 'jenkinsflow_test__retry_parallel_through_outer_level__j41_fail' failed but will be retried. Up to 1 more times in current flow" in sout, 'SOUT:' + sout
        assert "RETRY: job: 'jenkinsflow_test__retry_parallel_through_outer_level__j41_fail' failed but will be retried. Up to 2 more times through outer flow" in sout, 'SOUT:' + sout


def test_retry_serial_through_outer_level(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=2, serial=True)
        api.job('j31', max_fails=0, expect_invocations=1, expect_order=3, serial=True)
        api.job('j41', max_fails=0, expect_invocations=4, expect_order=3)
        api.job('j42_fail', max_fails=3, expect_invocations=4, expect_order=3, serial=True)
        api.job('j43', max_fails=0, expect_invocations=1, expect_order=3, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            ctrl1.invoke('j11')
            with ctrl1.serial(timeout=70) as ctrl2:
                ctrl2.invoke('j21')
                with ctrl2.parallel(timeout=70) as ctrl3:
                    ctrl3.invoke('j31')
                    with ctrl3.serial(timeout=70, max_tries=2) as ctrl4:
                        ctrl4.invoke('j41')
                        ctrl4.invoke('j42_fail')
                        ctrl4.invoke('j43')


def test_retry_mix(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11_fail', max_fails=1, expect_invocations=2, expect_order=1)
        api.job('j12', max_fails=0, expect_invocations=1, expect_order=2, serial=True)
        api.job('j21', max_fails=0, expect_invocations=1, expect_order=3, serial=True)
        api.job('j22_fail', max_fails=2, expect_invocations=3, expect_order=3)
        api.job('j31_fail', max_fails=3, expect_invocations=4, expect_order=3)
        api.job('j32', max_fails=0, expect_invocations=1, expect_order=3, serial=True)
        api.job('j23', max_fails=0, expect_invocations=1, expect_order=3)
        api.job('j13', max_fails=0, expect_invocations=1, expect_order=4, serial=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, max_tries=2) as ctrl1:
            ctrl1.invoke('j11_fail')
            ctrl1.invoke('j12')

            with ctrl1.parallel(timeout=70, max_tries=3) as ctrl2:
                ctrl2.invoke('j21')
                ctrl2.invoke('j22_fail')
                with ctrl2.serial(timeout=70, max_tries=2) as ctrl3:
                    ctrl3.invoke('j31_fail')
                    ctrl3.invoke('j32')
                ctrl2.invoke('j23')

            ctrl1.invoke('j13')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import pytest

from jenkinsflow.flow import serial
from .cfg import ApiType
from .framework import api_select


@pytest.mark.not_apis(ApiType.SCRIPT)
def test_env_variables(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, expect_order=1, print_env=True)
        api.job('j12', max_fails=0, expect_invocations=1, invocation_delay=1.0, exec_time=1.5, initial_buildno=7, expect_order=2, serial=True, print_env=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')
            ctrl1.invoke('j12')

        # TODO, check for env variables in log files



# -*- coding: utf-8 -*-

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import pytest

from jenkinsflow import flow
from jenkinsflow.api_base import AuthError


import demo_security as security

from .framework import api_select
from .cfg import ApiType


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_unicode_auth_latin1(api_type):
    with api_select.api(__file__, api_type, username="jenkinsflow_authtest1", password=r'abcæøåÆØÅ') as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, invocation_delay=1.0, expect_order=1)

        with flow.serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_unicode_auth_utf8(api_type):
    pytest.xfail("BasicAuth does not support unicode")  # TODO, if that gets supported, then enable this
    with api_select.api(__file__, api_type, username="jenkinsflow_authtest2", password=u'æøå¶¹²³³¼¢⅝÷«') as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, invocation_delay=1.0, expect_order=1)

        with flow.serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_unicode_auth_ascii(api_type):
    """Ensure that login without unicode works"""

    with api_select.api(__file__, api_type, username=security.username, password=security.password) as api:
        api.flow_job()
        api.job('j11', max_fails=0, expect_invocations=1, invocation_delay=1.0, expect_order=1)

        with flow.serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11')


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_unicode_auth_failed_login(api_type):
    """Ensure that login with bad password fails in a reasonable way."""

    with pytest.raises(AuthError) as exinfo:
        with api_select.api(__file__, api_type, username=security.username, password='notright') as api:
            api.flow_job()

    assert ("401 Client Error: Invalid password/token for user: jenkinsflow_jobrunner" in str(exinfo.value) or
            "401 Unauthorized user: 'jenkinsflow_jobrunner' for url: http://")



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function
from .jenkins_api import UnknownJobException

try:
    import tenjin
    from tenjin.helpers import *
    engine = tenjin.Engine()
except ImportError:  # pragma: no cover
    engine = None


def update_job(jenkins, job_name, config_xml, pre_delete=False, background=False):
    """Update or create a job in Jenkins.

    Args:
        jenkins (jenkins_api.Jenkins): Jenkins Api instance used for accessing jenkins.
        job_name (str): The name of the job.
        config_xml (str): The Jenkins job config xml.
        pre_delete (boolean): I the job exists it will be deleted and re-created instead of being updated.
    """

    try:
        if not pre_delete:
            jenkins.poll()
            job = jenkins.get_job(job_name)
            print('Updating job:', job_name)
            job.update_config(config_xml)
            return

        print('Deleting job:', job_name)
        jenkins.delete_job(job_name)
    except UnknownJobException:
        pass

    print('Creating job:', job_name)
    jenkins.create_job(job_name, config_xml)
    if not background:
        jenkins.poll()


def update_job_from_template(jenkins, job_name, config_xml_template, pre_delete=False, background=False, context=None):
    """Create or update a job based on a `Tenjin` http://www.kuwata-lab.com/tenjin/ config.xml template.

    Args:
        config_xml_template (str): Filename of tenjin config.xml template.
        context (dict): Values to be used for template substitution.

    See :py:func:`.update_job` for other parameters.
    """

    assert engine, "You must install tenjin (e.g.: pip install tenjin)"
    config_xml = engine.render(config_xml_template, context or {})
    update_job(jenkins, job_name, config_xml, pre_delete, background)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pytest import raises

from jenkinsflow import jenkins_api


def test_jenkins_api_init_api_no_password(api_type):
    with raises(Exception) as exinfo:
        jenkins_api.Jenkins("dummy", "dummy", username="hugo")
    assert "You must specify both username and password or neither" in str(exinfo.value)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pytest import raises

from jenkinsflow.flow import serial, JobNotIdleException
from .cfg import ApiType
from .framework import api_select
from .framework.utils import lines_in


def test_no_running_jobs(api_type, capsys):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=None, exec_time=50, invocation_delay=0, unknown_result=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke_unchecked('j1')

        sout, _ = capsys.readouterr()
        assert lines_in(api_type, sout, "unchecked job: 'jenkinsflow_test__no_running_jobs__j1' UNKNOWN - RUNNING")

        # Make sure job has actually started before entering new flow
        api.sleep(1)

        with raises(JobNotIdleException) as exinfo:
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
                ctrl1.invoke('j1')

        assert "job: 'jenkinsflow_test__no_running_jobs__j1' is in state RUNNING. It must be IDLE." in str(exinfo.value)


def test_no_running_jobs_unchecked(api_type, capsys):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=None, exec_time=50, invocation_delay=0, unknown_result=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke_unchecked('j1')

        sout, _ = capsys.readouterr()
        assert lines_in(api_type, sout, "unchecked job: 'jenkinsflow_test__no_running_jobs_unchecked__j1' UNKNOWN - RUNNING")

        api.sleep(1)

        with raises(JobNotIdleException) as exinfo:
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
                ctrl1.invoke_unchecked('j1')

        assert "unchecked job: 'jenkinsflow_test__no_running_jobs_unchecked__j1' is in state RUNNING. It must be IDLE." in str(exinfo.value)


def test_no_running_jobs_jobs_allowed(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        exp_invocations = 2 if api.api_type != ApiType.MOCK else 1
        unknown_result = False if api.api_type != ApiType.MOCK else True
        api.job('j1', max_fails=0, expect_invocations=exp_invocations, expect_order=None, exec_time=20,
                invocation_delay=0, unknown_result=unknown_result)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke_unchecked('j1')

        api.sleep(1)

        # TODO
        if api.api_type != ApiType.MOCK:
            with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, require_idle=False) as ctrl1:
                ctrl1.invoke('j1')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function

import sys, os, tempfile, time
from collections import OrderedDict

from .api_base import BuildResult, Progress, AuthError, UnknownJobException, ApiInvocationMixin
from .speed import Speed
from .rest_api_wrapper import ResourceNotFound, RequestsRestApi
from .jenkins_cli_protocol import CliProtocol


major_version = sys.version_info.major


if major_version < 3:
    from .rest_api_wrapper import ConnectionError


jenkins_cli_jar = 'jenkins-cli.jar'
hudson_cli_jar = 'hudson-cli.jar'

_superseded = -1
_dequeued = -2

_ct_url_enc = {'Content-Type': 'application/x-www-form-urlencoded'}


def _result_and_progress(build_dct):
    result = build_dct['result']
    progress = Progress.RUNNING if result is None else Progress.IDLE
    result = BuildResult.UNKNOWN if result is None else BuildResult[result]
    return (result, progress)


class Jenkins(Speed):
    """Optimized minimal set of methods needed for jenkinsflow to access Jenkins jobs.

    Args:
        direct_uri (str): Should be a non-proxied uri if possible (e.g. http://localhost:<port> if flow job is running on master)
            The public URI will be retrieved from Jenkins and used in output.
        job_prefix_filter (str): Jobs with names that don't start with this string, will be skpped when polling Jenkins.
            If you are using Hudson and have many jobs, it might be a good idea to enable Team support and create a job-runner user,
            which only has access to the jobs in the flow that it is executing. That way the job list will be filtered serverside.
        username (str): Name of user authorized to execute all jobs in flow.
        password (str): Password of user.
        invocation_class (class): Defaults to `Invocation`.
    """

    def __init__(self, direct_uri, job_prefix_filter=None, username=None, password=None, invocation_class=None, rest_access_provider=RequestsRestApi):
        if username or password:
            if not (username and password):
                raise Exception("You must specify both username and password or neither")
        self.rest_api = rest_access_provider(direct_uri, username, password)

        self.direct_uri = direct_uri
        self.username = username
        self.password = password
        self.invocation_class = invocation_class or Invocation
        self.job_prefix_filter = job_prefix_filter
        self._public_uri = None
        self.jobs = None
        self.queue_items = {}
        self.is_jenkins = None

    def get_content(self, url, **params):
        return self.rest_api.get_content(url, **params)

    def get_json(self, url="", **params):
        # Sometimes (but rarely) Jenkins will Abort the connection when jobs are being aborted!
        for ii in (1, 2, 3):
            try:
                return self.rest_api.get_json(url, **params)
            except ConnectionError as ex:
                if ii == 3:
                    raise
                print("WARNING: Retrying failed 'poll':", ex)
                time.sleep(0.1)

    def post(self, url, payload=None, headers=None, **params):
        return self.rest_api.post(url, payload, headers, **params)

    def headers(self):
        return self.rest_api.headers()

    @property
    def public_uri(self):
        if not self._public_uri:
            query = "primaryView[url]"
            dct = self.get_json(tree=query)
            self._public_uri = dct['primaryView']['url'].rstrip('/')
        return self._public_uri

    def _public_job_url(self, job_name):
        return self.public_uri + '/job/' + job_name

    def poll(self):
        # Determine whether we are talking to Jenkins or Hudson
        if self.is_jenkins is None:
            # TODO: A lot of Nonsense here because Hudson does not respond reliably
            for _ in (1, 2, 3):
                try:
                    head_response = self.headers()
                    if head_response.get("X-Jenkins"):
                        self.is_jenkins = True
                        break
                    if head_response.get("X-Hudson"):
                        self.is_jenkins = False
                        break
                except AuthError:
                    raise
                except Exception as ex:
                    head_response = "HEAD request failed: " + str(ex)
                time.sleep(0.1)
            else:
                raise Exception("Not connected to Jenkins or Hudson (expected X-Jenkins or X-Hudson header, got: " + repr(head_response))

        query = "jobs[name,lastBuild[number,result],queueItem[why],actions[parameterDefinitions[name,type]]],primaryView[url]"
        dct = self.get_json(tree=query)
        self._public_uri = dct['primaryView']['url'].rstrip('/')

        self.jobs = {}
        for job_dct in dct.get('jobs') or []:
            job_name = str(job_dct['name'])
            if self.job_prefix_filter and not job_name.startswith(self.job_prefix_filter):
                continue
            self.jobs[job_name] = ApiJob(self, job_dct, job_name)

    def quick_poll(self):
        query = "jobs[name,lastBuild[number,result],queueItem[why]]"
        dct = self.get_json(tree=query)

        for job_dct in dct.get('jobs') or []:
            job_name = str(job_dct['name'])
            if self.job_prefix_filter and not job_name.startswith(self.job_prefix_filter):
                continue
            job = self.jobs.get(job_name)
            if job:
                job.dct = job_dct
                continue

            # A new job was created while flow was running, get the remaining properties
            try:
                query = "lastBuild[number,result],queueItem[why],actions[parameterDefinitions[name,type]]"
                job_dct = self.get_json("/job/" + job_name, tree=query)
                job = ApiJob(self, job_dct, job_name)
                self.jobs[job_name] = job
            except ResourceNotFound:  # pragma: no cover
                # Ignore this, the job came and went
                pass

    def queue_poll(self):
        query = "items[task[name],id]"
        dct = self.get_json("/queue", tree=query)

        queue_items = {}
        for qi_dct in dct.get('items') or []:
            job_name = str(qi_dct['task']['name'])
            if self.job_prefix_filter and not job_name.startswith(self.job_prefix_filter):
                continue

            queue_items.setdefault(job_name, []).append(qi_dct['id'])
        self.queue_items = queue_items

    def get_job(self, name):
        try:
            return self.jobs[name]
        except KeyError:
            raise UnknownJobException(self._public_job_url(name))

    def create_job(self, job_name, config_xml):
        self.post('/createItem', name=job_name,
                  headers={'Content-Type': 'application/xml header;charset=utf-8'},
                  payload=config_xml if major_version < 3 else config_xml.encode('utf-8'))

    def delete_job(self, job_name):
        try:
            self.post('/job/' + job_name + '/doDelete')
        except ResourceNotFound as ex:
            # TODO: Check error
            raise UnknownJobException(self._public_job_url(job_name), ex)

    def set_build_description(self, description, replace=False, separator='\n', job_name=None, build_number=None):
        """Utility to set/append build description

        Args:
            description (str):  The description to set on the build
            replace (bool):     If True, replace existing description, if any, instead of appending to it
            separator (str):    A separator to insert between any existing description and the new :py:obj:`description` if :py:obj:`replace` is False.
            job_name (str):     Name of the Jenkins job
            build_number (int): The build number for which to set the description
        """
        self.poll()

        if job_name is None:
            job_name = os.environ['JOB_NAME']

        if build_number is None:
            build_number = int(os.environ['BUILD_NUMBER'])

        job_path = "/job/" + job_name
        build_url = job_path + '/' + str(build_number)
        try:
            if not replace:
                dct = self.get_json(build_url, tree="description")
                existing_description = dct['description']
                if existing_description:
                    description = existing_description + separator + description

            self.post(build_url + '/submitDescription', headers=_ct_url_enc, payload={'description': description})
        except ResourceNotFound as ex:
            raise Exception("Build not found " + repr(build_url), ex)

    def _download_cli(self, cli_jar):
        public_uri = self.public_uri.rstrip('/') + '/'
        direct_uri = self.direct_uri.rstrip('/') + '/'

        path = 'jnlpJars/' + cli_jar
        public_cli_url = public_uri + path
        if direct_uri != public_uri:
            download_cli_url = direct_uri + path
            print("INFO: Downloading cli: '{public_cli_url}' (using direct url: '{direct_cli_url}')".format(
                public_cli_url=public_cli_url, direct_cli_url=download_cli_url))
        else:
            download_cli_url = public_cli_url
            print("INFO: Downloading cli: '{download_cli_url}'".format(download_cli_url=download_cli_url))

        with open(cli_jar, 'w' if major_version < 3 else 'w+b') as ff:
            ff.write(self.get_content('/' + path))
        print("INFO: Download finished:", repr(cli_jar))

    def set_build_result(self, result, java='java', cli_call=False, protocol=CliProtocol.remoting):
        """Change the result of a Jenkins job.

        DEPRECATED - You should use the shell step exit code to determine the job result.

        Note: set_build_result can only be done from within the job, not after the job has finished.
        Note: Only available if URL is set in `Jenkins <http://jenkins-ci.org/>`_ system configuration.

        This command uses the Jenkins `cli` to change the result. It requires a java executable to run the cli.
        Note: In some versions of Jenkins the `cli` is broken, it has no manifest file! This is the case for
        e.g. 1.625.1, the version installed on Fedora 23 using `dnf` at the time of Fedora 23 release.

        Args:
            result (str): The result to set. Should probably be 'unstable'
            java (str): Alternative `java` executable. Use this if you don't wish to use the java in the PATH.
            protocol (CliProtocol): Defaults to 'remoting' for backwards compatibility. See: https://jenkins.io/doc/book/managing/cli/.
        """

        print("INFO: Setting job result to", repr(result))
        cli_jar = jenkins_cli_jar if self.is_jenkins else hudson_cli_jar
        protocol_option = ['-' + protocol.name] if protocol else []

        if major_version < 3:
            import subprocess32 as subprocess
        else:
            import subprocess

        def set_res():
            command = [java, '-jar', cli_jar, '-s', self.direct_uri] + protocol_option + ['set-build-result', result]
            if self.username:
                fname = None
                try:
                    fhandle, fname = tempfile.mkstemp()
                    fhandle = os.fdopen(fhandle, 'w')
                    fhandle.write(self.password)
                    fhandle.close()
                    subprocess.check_call(command + ['--username', self.username, '--password-file', fname])
                finally:
                    try:
                        os.remove(fname)
                        fhandle.close()
                    except IOError:  # pragma: no cover
                        pass
            else:
                subprocess.check_call(command)

        try:
            # If cli_jar is already present attempt to use it
            set_res()
        except subprocess.CalledProcessError:
            # We failed for some reason, try again with updated cli_jar
            self._download_cli(cli_jar)
            set_res()


class ApiJob(object):
    def __init__(self, jenkins, dct, name):
        self.jenkins = jenkins
        self.dct = dct.copy()
        self.name = name
        self.public_uri = self.jenkins._public_job_url(self.name)  # pylint: disable=protected-access

        actions = self.dct.get('actions') or []
        self._path = "/job/" + self.name
        for action in actions:
            if action is None:
                continue
            if action.get('parameterDefinitions'):
                self._build_trigger_path = self._path + "/buildWithParameters"
                break
        else:
            self._build_trigger_path = self._path + "/build"
        self.old_build_number = None
        self._invocations = OrderedDict()
        self.queued_why = None

    def invoke(self, securitytoken, build_params, cause, description):
        try:
            if cause:
                build_params = build_params or {}
                build_params['cause'] = cause
            headers = _ct_url_enc if build_params else None
            params = {}
            if securitytoken:
                params['token'] = securitytoken
            response = self.jenkins.post(self._build_trigger_path, headers=headers, payload=build_params, **params)
        except ResourceNotFound as ex:
            raise UnknownJobException(self.jenkins._public_job_url(self.name), ex)  # pylint: disable=protected-access

        location = response.headers['location'][len(self.jenkins.direct_uri):-1]
        old_inv = self._invocations.get(location)
        if old_inv:
            old_inv.build_number = _superseded
        inv = self.jenkins.invocation_class(self, location, description)
        self._invocations[location] = inv
        return inv

    def poll(self):
        for invocation in self._invocations.values():
            if not invocation.build_number:
                # Hudson does not return queue item from invoke, instead it returns the job URL :(
                query = "executable[number],why" if self.jenkins.is_jenkins else "queueItem[why],lastBuild[number]"
                dct = self.jenkins.get_json(invocation.queued_item_path, tree=query)

                if self.jenkins.is_jenkins:
                    executable = dct.get('executable')
                    if executable:
                        invocation.build_number = executable['number']
                        invocation.queued_why = None
                        invocation.set_description()
                    else:
                        invocation.queued_why = dct['why']
                        # If we still have invocations in the queue, wait until next poll to query again
                        break
                else:  # Hudson
                    # Note, this is not guaranteed to be correct in case of simultaneously running flows!
                    # Should handle multiple invocations in same flow
                    qi = dct.get('queueItem')
                    if qi:
                        invocation.queued_why = qi['why']

                    last_build = dct.get('lastBuild')
                    if last_build:
                        last_build_number = last_build['number']
                        if last_build_number > self.old_build_number:
                            invocation.build_number = last_build['number']
                            self.old_build_number = invocation.build_number
                            invocation.set_description()
                        else:
                            break

    def job_status(self):
        """Result, progress and latest buildnumber info for the JOB, NOT the invocation

        Return (result, progress_info, latest_build_number) (str, str, int or None):
            If there is no finished build, result will be BuildResult.UNKNOWN and latest_build_number will be None
        """
        progress = None

        qi = self.dct['queueItem']
        if qi:
            progress = Progress.QUEUED
            self.queued_why = qi['why']

        dct = self.dct.get('lastBuild')
        if dct:
            self.old_build_number = dct['number']
            result, latest_progress = _result_and_progress(dct)
            return (result, progress or latest_progress, self.old_build_number)

        return (BuildResult.UNKNOWN, progress or Progress.IDLE, None)

    def stop_all(self):
        # First remove pending builds from queue
        queue_item_ids = self.jenkins.queue_items.get(self.name) or []
        for qid in queue_item_ids:
            try:
                self.jenkins.post('/queue/cancelItem', id=repr(qid))
            except ResourceNotFound:
                # Job is no longer queued, so just ignore
                # NOTE: bug https://issues.jenkins-ci.org/browse/JENKINS-21311 also brings us here!
                pass

        # Abort running builds
        query = "builds[number,result]"
        dct = self.jenkins.get_json("/job/" + self.name, tree=query)
        for build in dct['builds']:
            _result, progress = _result_and_progress(build)
            if progress != Progress.IDLE:
                build_number = build['number']
                try:
                    self.jenkins.post(self._path + '/' + repr(build_number) + '/stop')
                except ResourceNotFound:  # pragma: no cover
                    # Build was deleted, just ignore
                    pass

    def update_config(self, config_xml):
        self.jenkins.post("/job/" + self.name + "/config.xml",
                          headers={'Content-Type': 'application/xml header;charset=utf-8'},
                          payload=config_xml if major_version < 3 else config_xml.encode('utf-8'))

    def disable(self):
        try:
            self.jenkins.post(self._path + '/disable')
        except ResourceNotFound as ex:
            raise UnknownJobException(self.jenkins._public_job_url(self.name), ex)  # pylint: disable=protected-access

    def enable(self):
        try:
            self.jenkins.post(self._path + '/enable')
        except ResourceNotFound as ex:
            raise UnknownJobException(self.jenkins._public_job_url(self.name), ex)  # pylint: disable=protected-access

    def __repr__(self):
        return str(dict(name=self.name, dct=self.dct))


class Invocation(ApiInvocationMixin):
    def __init__(self, job, queued_item_path, description):
        self.job = job
        self.queued_item_path = queued_item_path
        self.description = description
        self.build_number = None
        self.queued_why = None

    def __repr__(self):
        return 'Invocation: ' + repr(self.queued_item_path) + ' ' + repr(self.build_number) + ' ' + repr(self.queued_why)

    def status(self):
        """Result and Progress info for the invocation

        Return (result, progress_info) (str, str):
            If the build has not started or has not finished running, result will be BuildResult.UNKNOWN
        """

        if self.build_number is None:
            return (BuildResult.UNKNOWN, Progress.QUEUED)

        if self.build_number == _superseded:
            return (BuildResult.SUPERSEDED, Progress.IDLE)

        if self.build_number == _dequeued:
            return (BuildResult.DEQUEUED, Progress.IDLE)

        # It seems that even after the executor has been assigned a number in the queue item, the lastBuild might not yet exist
        dct = self.job.dct.get('lastBuild')
        last_number = dct['number'] if dct else None
        if last_number is None:
            return (BuildResult.UNKNOWN, Progress.QUEUED)

        if last_number == self.build_number:
            return _result_and_progress(dct)

        if last_number < self.build_number:
            # TODO: Why does this happen?
            pass  # pragma: no cover

        # Latest build is not ours, get the correct build
        query = "builds[number,result]"
        dct = self.job.jenkins.get_json("/job/" + self.job.name, tree=query)
        for build in dct['builds']:
            if build['number'] == self.build_number:
                return _result_and_progress(build)

        raise Exception("Build deleted while flow running? This may happen if you invoke more builds than the job is configured to keep. " + repr(self))

    def set_description(self):
        """Sets the build description"""
        if not self.description:
            return

        build_url = self.job._path + '/' + repr(self.build_number)
        try:
            self.job.jenkins.post(build_url + '/submitDescription', headers=_ct_url_enc, payload={'description': self.description})
        except ResourceNotFound as ex:
            raise Exception("Build deleted while flow running? " + repr(build_url), ex)

    def stop(self, dequeue):
        try:
            if self.build_number is not None and self.build_number >= 0 and not dequeue:
                # Job has started
                self.job.jenkins.post(self.job._path + '/' + repr(self.build_number) + '/stop')
                return

            if self.build_number is None and dequeue:
                # Job is queued
                qid = self.queued_item_path.strip('/').split('/')[2]
                self.job.jenkins.post('/queue/cancelItem', id=qid)
                self.build_number = _dequeued
        except ResourceNotFound as ex:  # pragma: no cover
            # Job is no longer queued or running, except that it may have just changed from queued to running
            # We leave it up to the flow logic to handle that
            # NOTE: bug https://issues.jenkins-ci.org/browse/JENKINS-21311 also brings us here!
            pass



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from os.path import join as jp

# Note: duplicated in tmp_install.sh and INSTALL.md
test_tmp_dir = "/tmp/jenkinsflow-test"
pseudo_install_dir = jp(test_tmp_dir, "jenkinsflow")
flow_graph_root_dir = jp(test_tmp_dir, "graphs")
job_script_dir = jp(test_tmp_dir, 'job')



"""Move time and sleep to api class, so that the MockApi can speed up time."""


import time as real_time


class Speed(object):
    speedup = 1

    def time(self):
        return real_time.time()

    def sleep(self, seconds):
        return real_time.sleep(seconds)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pytest import raises

from jenkinsflow.flow import parallel, serial, FlowScopeException
from .framework import api_select


def test_flow_scope_job(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=0, expect_order=None)
        api.job('j2', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(FlowScopeException):
            with serial(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                pass
            ctrl1.invoke('j1')

        with raises(FlowScopeException):
            with parallel(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                pass
            ctrl1.invoke('j1')

        with raises(FlowScopeException):
            with serial(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                with ctrl1.parallel() as ctrl2:
                    pass
                ctrl2.invoke('j2')

        with raises(FlowScopeException):
            with parallel(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                with ctrl1.serial() as ctrl2:
                    ctrl2.invoke('j1')
                ctrl2.invoke('j2')


def test_flow_scope_serial(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()

        with raises(FlowScopeException):
            with serial(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                pass
            with ctrl1.serial(1):
                pass

        with raises(FlowScopeException):
            with serial(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                pass
            with ctrl1.parallel(1):
                pass

        with raises(FlowScopeException):
            with serial(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                with ctrl1.serial() as ctrl2:
                    ctrl2.serial(1)
                ctrl2.serial(1)

        with raises(FlowScopeException):
            with serial(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                with ctrl1.serial() as ctrl2:
                    ctrl1.serial(1)


def test_flow_scope_parallel(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()

        with raises(FlowScopeException):
            with parallel(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                pass
            with ctrl1.parallel(1):
                pass

        with raises(FlowScopeException):
            with parallel(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                pass
            with ctrl1.serial(1):
                pass

        with raises(FlowScopeException):
            with parallel(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                with ctrl1.parallel() as ctrl2:
                    ctrl2.parallel(1)
                ctrl2.parallel(1)

        with raises(FlowScopeException):
            with parallel(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                with ctrl1.parallel() as ctrl2:
                    ctrl1.parallel(1)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

# NOTE: the tests here all raise exceptions because they can not really be run outside of a jenkinsjob
# TODO: Test that the script actually does what expected! The test here just assure that the script can be run :(

import sys, os, re
major_version = sys.version_info.major
if major_version < 3:
    import subprocess32 as subprocess
    import urllib2 as urllib
else:
    import subprocess, urllib

from os.path import join as jp

import pytest
from pytest import raises

from jenkinsflow.flow import serial, Propagation, BuildResult
from jenkinsflow.cli.cli import cli

from .framework import api_select
from .framework.utils import lines_in, jenkins_cli_jar, hudson_cli_jar
from . import cfg as test_cfg
from .cfg import ApiType

from demo_security import username, password

here = os.path.abspath(os.path.dirname(__file__))


def _download_cli_jar(cli_jar, direct_base_url, public_base_url):
    if major_version < 3:
        from urllib2 import urlopen
        save_mode = 'w'
    else:
        from urllib.request import urlopen
        save_mode = 'w+b'

    path = '/jnlpJars/' + cli_jar
    if direct_base_url and direct_base_url != public_base_url:
        download_cli_url = direct_base_url + path
        print("Prereqs: Downloading cli:", repr(public_base_url + path), "(using direct url:", download_cli_url + ')')
    else:
        download_cli_url = public_base_url + path
        print("Prereqs: Downloading cli:", repr(download_cli_url))

    response = urlopen(download_cli_url)
    with open(cli_jar, save_mode) as ff:
        ff.write(response.read())
    print("Prereqs: Download finished:", repr(cli_jar))


def pre_existing_cli_jar(api_type):
    if api_type != ApiType.JENKINS:
        return

    public_base_url = os.environ.get('HUDSON_URL')
    cli_jar = hudson_cli_jar

    if public_base_url is None:
        public_base_url = test_cfg.public_url(api_type)
        cli_jar = jenkins_cli_jar

    if not os.path.exists(cli_jar):
        _download_cli_jar(cli_jar, test_cfg.direct_url(api_type), public_base_url)


def no_pre_existing_cli_jar(api_type):
    if api_type == ApiType.SCRIPT:
        return

    if os.path.exists(jenkins_cli_jar):
        os.remove(jenkins_cli_jar)
    if os.path.exists(hudson_cli_jar):
        os.remove(hudson_cli_jar)


_setting_job_result_msg = "INFO: Setting job result to 'unstable'"


def _download_same_url(download_url):
    return "^INFO: Downloading cli: {download_url!r}".format(download_url=download_url)


def _download_different_urls(api_type):
    return "^INFO: Downloading cli: {public_url!r} (using direct url: {direct_url!r})".format(
        public_url=test_cfg.proxied_public_cli_url(api_type), direct_url=test_cfg.direct_cli_url(api_type))


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_result_no_cli_jar(api_type, fake_java, env_base_url, capfd):
    with api_select.api(__file__, api_type, fake_public_uri=test_cfg.direct_url(api_type)) as api:
        no_pre_existing_cli_jar(api_type)
        api.flow_job()
        api.job('j1_fail', max_fails=1, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, username=username, password=password, job_name_prefix=api.job_name_prefix, report_interval=3,
                    propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl1:
            ctrl1.invoke('j1_fail')

    sout, _ = capfd.readouterr()

    expected = [_setting_job_result_msg]
    if api.api_type != ApiType.SCRIPT:
        expected.append(_download_same_url(test_cfg.direct_cli_url(api_type)))
        assert '/jnlpJars/' in sout

    assert lines_in(
        api_type, sout,
        *expected
    )


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_result_no_cli_jar_env_base_url_trailing_slash(api_type, fake_java, env_base_url_trailing_slash, capfd):
    with api_select.api(__file__, api_type, fake_public_uri=test_cfg.direct_url(api_type) + '/') as api:
        no_pre_existing_cli_jar(api_type)
        api.flow_job()
        api.job('j1_fail', max_fails=1, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, username=username, password=password, job_name_prefix=api.job_name_prefix, report_interval=3,
                    propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl1:
            ctrl1.invoke('j1_fail')

    sout, _ = capfd.readouterr()

    expected = [_setting_job_result_msg]
    if api.api_type != ApiType.SCRIPT:
        expected.append(_download_same_url(test_cfg.direct_cli_url(api_type)))
        assert '/jnlpJars/' in sout

    assert lines_in(
        api_type, sout,
        *expected
    )


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_result_no_cli_jar_env_base_url_trailing_slashes(api_type, fake_java, capfd):
    with api_select.api(__file__, api_type, fake_public_uri=test_cfg.direct_url(api_type) + '//') as api:
        no_pre_existing_cli_jar(api_type)
        api.flow_job()
        api.job('j1_fail', max_fails=1, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, username=username, password=password, job_name_prefix=api.job_name_prefix, report_interval=3,
                    propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl1:
            ctrl1.invoke('j1_fail')

    sout, _ = capfd.readouterr()

    expected = [_setting_job_result_msg]
    if api.api_type != ApiType.SCRIPT:
        expected.append(_download_same_url(test_cfg.direct_cli_url(api_type)))
        assert '/jnlpJars/' in sout

    assert lines_in(
        api_type, sout,
        *expected
    )


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_result(api_type, fake_java):
    with api_select.api(__file__, api_type) as api:
        pre_existing_cli_jar(api_type)
        api.flow_job()
        api.job('j1_fail', max_fails=1, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, username=username, password=password, job_name_prefix=api.job_name_prefix, report_interval=3,
                    propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl1:
            ctrl1.invoke('j1_fail')


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_result_direct_url(api_type, fake_java):
    with api_select.api(__file__, api_type) as api:
        pre_existing_cli_jar(api_type)
        api.flow_job()
        api.job('j1_fail', max_fails=1, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, username=username, password=password, job_name_prefix=api.job_name_prefix, report_interval=3,
                    propagation=Propagation.FAILURE_TO_UNSTABLE, direct_url=test_cfg.direct_url(api_type)) as ctrl1:
            ctrl1.invoke('j1_fail')


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_result_no_cli_jar_env_base_url_eq_direct_url(api_type, fake_java, env_base_url, capfd):
    with api_select.api(__file__, api_type, fake_public_uri=test_cfg.direct_url(api_type)) as api:
        no_pre_existing_cli_jar(api_type)
        api.flow_job()
        api.job('j1_fail', max_fails=1, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, username=username, password=password, job_name_prefix=api.job_name_prefix, report_interval=3,
                    propagation=Propagation.FAILURE_TO_UNSTABLE, direct_url=test_cfg.public_url(api_type)) as ctrl1:
            ctrl1.invoke('j1_fail')

    sout, _ = capfd.readouterr()

    expected = [_setting_job_result_msg]
    if api.api_type != ApiType.SCRIPT:
        expected.append(re.compile(_download_same_url(test_cfg.direct_cli_url(api_type)) + '$'))
        expected.append("^INFO: Download finished: ")
        assert '/jnlpJars/' in sout

    assert lines_in(
        api_type, sout,
        *expected
    )


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_result_direct_url_trailing_slash(api_type, fake_java, env_base_url, capfd):
    with api_select.api(__file__, api_type) as api:
        pre_existing_cli_jar(api_type)
        api.flow_job()
        api.job('j1_fail', max_fails=1, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, username=username, password=password, job_name_prefix=api.job_name_prefix, report_interval=3,
                    propagation=Propagation.FAILURE_TO_UNSTABLE, direct_url=test_cfg.direct_url(api_type) + '/') as ctrl1:
            ctrl1.invoke('j1_fail')

    sout, _ = capfd.readouterr()
    assert lines_in(
        api_type, sout,
        _setting_job_result_msg
    )


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_result_direct_url_different_from_proxied_url(api_type, fake_java, capfd):
    with api_select.api(__file__, api_type, fake_public_uri=test_cfg.proxied_public_url) as api:
        no_pre_existing_cli_jar(api_type)
        api.flow_job()
        api.job('j1_fail', max_fails=1, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, username=username, password=password, job_name_prefix=api.job_name_prefix, report_interval=3,
                    propagation=Propagation.FAILURE_TO_UNSTABLE, direct_url=test_cfg.direct_url(api_type) + '/') as ctrl1:
            ctrl1.invoke('j1_fail')

    sout, _ = capfd.readouterr()

    expected = [_setting_job_result_msg]
    if api.api_type != ApiType.SCRIPT:
        expected.append(_download_different_urls(api_type))

    assert lines_in(
        api_type, sout,
        *expected
    )


@pytest.mark.not_apis(ApiType.MOCK)
def test_set_build_result_no_auth(api_type, fake_java, env_base_url):
    with api_select.api(__file__, api_type) as api:
        pre_existing_cli_jar(api_type)
        api.flow_job()
        api.job('j1_fail', max_fails=1, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=3, propagation=Propagation.FAILURE_TO_UNSTABLE) as ctrl1:
            ctrl1.invoke('j1_fail')


@pytest.mark.apis(ApiType.SCRIPT)
def test_set_build_result_unstable_script_api(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j11_unstable', max_fails=0, expect_invocations=1, expect_order=1, final_result='UNSTABLE', final_result_use_cli=True)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j11_unstable')

        assert ctrl1.result == BuildResult.UNSTABLE


@pytest.mark.not_apis(ApiType.SCRIPT)  # JENKINS_URL is always set for script_api
def test_set_build_result_no_jenkinsurl(api_type, fake_java, env_no_base_url, cli_runner):
    with api_select.api(__file__, api_type):
        pre_existing_cli_jar(api_type)
        result = cli_runner.invoke(cli, ['set_build_result'])

    assert result.exit_code != 0
    assert result.exception

    assert lines_in(
        api_type, str(result.exception),
        "Could not get env variable JENKINS_URL or HUDSON_URL. You must set 'Jenkins Location' in Jenkins setup for JENKINS_URL to be exported. You must set 'Hudson URL' in Hudson setup for HUDSON_URL to be exported."
    )


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_set_build_result_call_cli_direct_url_trailing_slash(api_type, fake_java, env_base_url, cli_runner):
    with api_select.api(__file__, api_type) as api:
        pre_existing_cli_jar(api_type)
        base_url = test_cfg.direct_url(api_type) + '/'

        api.flow_job()
        api.job('j1_unstable', max_fails=0, expect_invocations=1, expect_order=1,
                final_result='UNSTABLE', final_result_use_cli=True, set_build_result_use_url=base_url)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j1_unstable')

        assert ctrl1.result == BuildResult.UNSTABLE


@pytest.mark.apis(ApiType.SCRIPT)
def test_set_build_result_call_cli_direct_url_trailing_slash_script_api(api_type, env_base_url, cli_runner):
    with api_select.api(__file__, api_type):
        pre_existing_cli_jar(api_type)
        base_url = test_cfg.direct_url(api_type) + '/'
        result = cli_runner.invoke(cli, ['set_build_result', '--direct-url', base_url])

    assert result.exit_code != 0
    assert result.exception

    assert lines_in(
        api_type, str(result.exception),
        "Could not get EXECUTOR_NUMBER from env. 'set_build_result' must be invoked from within a running job"
    )


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_set_build_result_call_main_direct_url_no_trailing_slash(api_type, fake_java, env_base_url, cli_runner):
    with api_select.api(__file__, api_type) as api:
        pre_existing_cli_jar(api_type)
        base_url = test_cfg.direct_url(api_type).rstrip('/')

        api.flow_job()
        api.job('j1_unstable', max_fails=0, expect_invocations=1, expect_order=1,
                final_result='UNSTABLE', final_result_use_cli=True, set_build_result_use_url=base_url)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke('j1_unstable')

        assert ctrl1.result == BuildResult.UNSTABLE


def test_set_build_result_call_script_help(capfd):
    # Invoke this in a subprocess to ensure that calling the script works
    # This will not give coverage as it not not traced through the subprocess call
    rc = subprocess.call([sys.executable, jp(here, '../cli/cli.py'), 'set_build_result', '--help'])
    assert rc == 0

    sout, _ = capfd.readouterr()
    assert '--result' in sout
    assert '--direct-url' in sout
    assert '--username' in sout
    assert '--password' in sout
    assert '--java' in sout



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function

import click

from ..utils import set_build_description as usbd


@click.command()
@click.option('--description', help="The description to set on the build")
@click.option('--replace/--no-replace', default=False, help="Replace existing description, if any, instead of appending.")
@click.option('--separator', default='\n', help="A separator to insert between any existing description and the new 'description' if 'replace' is not specified.")
@click.option('--username', help="User Name for Jenkin authentication with secured Jenkins")
@click.option('--password', help="Password of Jenkins User")
@click.option('--job-name', help='Job Name', envvar='JOB_NAME')
@click.option('--build-number', help="Build Number", type=click.INT, envvar='BUILD_NUMBER')
@click.option(
    '--direct-url',
    default=None,
    help="Jenkins URL - preferably non-proxied. If not specified, the value of JENKINS_URL or HUDSON_URL environment variables will be used.")
def set_build_description(description, replace, separator, username, password, job_name, build_number, direct_url):
    """Utility to set/append build description on a job build."""
    # %(file)s --job-name <job_name> --build-number <build_number> --description <description> [--direct-url <direct_url>] [--replace | --separator <separator>] [(--username <user_name> --password <password>)]

    usbd.set_build_description(description, replace, separator, username, password, job_name, build_number, direct_url)



#!/usr/bin/env python

"""
Test jenkinsflow.
"""

from __future__ import print_function

import sys, os, getpass, shutil, copy, errno
major_version = sys.version_info.major
if major_version < 3:
    import subprocess32 as subprocess
else:
    import subprocess
from os.path import join as jp

import click
import tenjin
from tenjin.helpers import *

try:
    import pytest
except ImportError:
    print("See setup.py for test requirements, or use 'python setup.py test'", file=sys.stderr)
    raise

here = os.path.abspath(os.path.dirname(__file__))
top_dir = os.path.dirname(here)

extra_sys_path = [os.path.normpath(path) for path in [here, jp(top_dir, '..'), jp(top_dir, 'demo'), jp(top_dir, 'demo/jobs')]]
sys.path = extra_sys_path + sys.path
os.environ['PYTHONPATH'] = ':'.join(extra_sys_path)

from jenkinsflow.test.framework import config
from jenkinsflow.test import cfg as test_cfg
from jenkinsflow.test.cfg import ApiType


def run_tests(parallel, api_types, args, coverage=True, mock_speedup=1):
    args = copy.copy(args)

    test_cfg.select_speedup(mock_speedup)

    if coverage:
        engine = tenjin.Engine()

        if len(api_types) == 3:
            fail_under = 95
        elif ApiType.JENKINS in api_types:
            fail_under = 94
        elif ApiType.MOCK in api_types and ApiType.SCRIPT in api_types:
            fail_under = 90
        elif ApiType.MOCK in api_types:
            fail_under = 88
        else:
            fail_under = 86

        # Note: cov_rc_file_name hardcoded in .travis.yml
        cov_rc_file_name = jp(here, '.coverage_rc_' +  '_'.join(api_type.name.lower() for api_type in api_types))
        with open(cov_rc_file_name, 'w') as cov_rc_file:
            context = dict(api_types=api_types, top_dir=top_dir, major_version=major_version, fail_under=fail_under)
            cov_rc_file.write(engine.render(jp(here, "coverage_rc.tenjin"), context))
            args.extend(['--cov=' + top_dir, '--cov-report=term-missing', '--cov-config=' + cov_rc_file_name])

    if api_types != [ApiType.MOCK]:
        # Note: 'boxed' is required for the kill/abort_current test not to abort other tests
        args.append('--boxed')

        if parallel:
            args.extend(['-n', '16'])

    print('pytest.main', args)
    rc = pytest.main(args)
    if rc:
        raise Exception("pytest {args} failed with code {rc}".format(args=args, rc=rc))


def start_msg(*msg):
    print("\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n", *msg)


def cli(mock_speedup=1000,
        direct_url=test_cfg.direct_url(test_cfg.ApiType.JENKINS),
        apis=None,
        pytest_args=None,
        job_delete=False,
        job_load=True,
        testfile=None):
    """
    Test jenkinsflow.
    First runs all tests mocked in hyperspeed, then runs against Jenkins, using jenkins_api, then run script_api jobs.

    Normally jobs will be run in parallel, specifying --job-delete disables this.
    The default options assumes that re-loading without deletions generates correct job config.
    Tests that require jobs to be deleted/non-existing will delete the jobs, regardless of the --job-delete option.

    [TESTFILE]... File names to pass to py.test
    """

    os.environ[test_cfg.DIRECT_URL_NAME] = direct_url
    os.environ[test_cfg.SKIP_JOB_DELETE_NAME] = 'false' if job_delete else 'true'
    os.environ[test_cfg.SKIP_JOB_LOAD_NAME] = 'false' if job_load else 'true'

    args = ['--capture=sys', '--instafail']

    if apis is None:
        api_types = [ApiType.MOCK, ApiType.SCRIPT, ApiType.JENKINS]
    else:
        api_types = [ApiType[api_name.strip().upper()] for api_name in apis.split(',')]
    args.extend(['-k', ' or '.join([apit.name for apit in api_types])])

    rc = 0
    target_dir = "/tmp/jenkinsflow-test/jenkinsflow"
    try:
        os.makedirs(target_dir)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise

    if api_types != [ApiType.MOCK]:
        print("Creating temporary test installation in", repr(config.pseudo_install_dir), "to make files available to Jenkins.")
        install_script = jp(here, 'tmp_install.sh')
        rc = subprocess.call([install_script, target_dir])
        if rc:
            print("Failed test installation to", repr(config.pseudo_install_dir), "Install script is:", repr(install_script), file=sys.stderr)
            print("Warning: Some tests will fail!", file=sys.stderr)

    cov_file = ".coverage"
    for cov_file in jp(here, cov_file), jp(top_dir, cov_file):
        if os.path.exists(cov_file):
            os.remove(cov_file)

    print("\nRunning tests")
    try:
        if pytest_args or testfile:
            coverage = False
            args.extend(pytest_args.split(' ') + list(testfile) if pytest_args else list(testfile))
        else:
            coverage = True
            args.append('--ff')

        hudson = os.environ.get('HUDSON_URL')
        if hudson:
            print("Disabling parallel run, Hudson can't handle it :(")
        parallel = test_cfg.skip_job_load() or test_cfg.skip_job_delete() and not hudson
        run_tests(parallel, api_types, args, coverage, mock_speedup)

        # start_msg("Testing setup.py")
        # user = getpass.getuser()
        # install_prefix = '/tmp/' + user
        # tmp_packages_dir = install_prefix + '/lib/python{major}.{minor}/site-packages'.format(major=major_version, minor=sys.version_info.minor)
        # os.environ['PYTHONPATH'] = tmp_packages_dir
        # if os.path.exists(tmp_packages_dir):
        #     shutil.rmtree(tmp_packages_dir)
        # os.makedirs(tmp_packages_dir)
        #
        # os.chdir(top_dir)
        # subprocess.check_call([sys.executable, jp(top_dir, 'setup.py'), 'install', '--prefix', install_prefix])
        # shutil.rmtree(jp(top_dir, 'build'))

        if not testfile and os.environ.get('CI', 'false').lower() != 'true':
            # This is automatically tested by readdthedocs, so no need to test on Travis
            start_msg("Testing documentation generation")

            os.chdir(jp(top_dir, 'doc/source'))
            del os.environ['PYTHONPATH']
            subprocess.check_call(['make', 'html'])
    except Exception as ex:
        print('*** ERROR: There were errors! Check output! ***', repr(ex), file=sys.stderr)
        raise

    sys.exit(rc)


@click.command()
@click.option('--mock-speedup', '-s', help="Time speedup when running mocked tests.", default=1000)
@click.option('--direct-url', help="Direct Jenkins URL. Must be different from the URL set in Jenkins (and preferably non proxied)",
              default=test_cfg.direct_url(test_cfg.ApiType.JENKINS))
@click.option('--apis', help="Select which api(s) to use/test. Comma separated list. Possible values: 'jenkins, 'script', 'mock'. Default is all.", default=None)
@click.option('--pytest-args', help="py.test arguments.")
@click.option('--job-delete/--no-job-delete', help="Delete and re-load jobs into Jenkins. Default is --no-job-delete.", default=False)
@click.option('--job-load/--no-job-load', help="Load jobs into Jenkins (skipping job load assumes all jobs already loaded and up to date). Deafult is --job-load.", default=True)
@click.argument('testfile', nargs=-1, type=click.Path(exists=True, readable=True))
def _cli(mock_speedup, direct_url, apis, pytest_args, job_delete, job_load, testfile):
    cli(mock_speedup, direct_url, apis, pytest_args, job_delete, job_load, testfile)


if __name__ == '__main__':
    _cli()  # pylint: disable=no-value-for-parameter



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os, re, tempfile
from os.path import join as jp
from functools import partial

import pytest

from jenkinsflow.test import cfg as test_cfg
from jenkinsflow.test.cfg import ApiType
from .config import flow_graph_root_dir
from .lines_in import lines_in as generic_lines_in


_log_file_prefix = tempfile.gettempdir() + '/jenkinsflow/'


def _script_api_log_file(job_name):
    return _log_file_prefix + job_name + '.log'


def console_url(api, job_name, num):
    if api.api_type == ApiType.SCRIPT:
        return _script_api_log_file(job_name)
    return 'http://x.x/job/' + job_name + '/' + (str(num) + "/console" if api.api_type == ApiType.MOCK else "")


def result_msg(api, job_name, num=None):
    if api.api_type == ApiType.SCRIPT:
        return repr(job_name) + " - build: " + _script_api_log_file(job_name)
    return repr(job_name) + " - build: http://x.x/job/" + job_name + "/" + (str(num) + "/console" if api.api_type == ApiType.MOCK and num else "")


def build_started_msg(api, job_name, num, invocation_number=0):
    return "Build started: " + repr(job_name) + ((' Invocation-' + repr(invocation_number)) if invocation_number else '') + " - " + console_url(api, job_name, num)


def kill_current_msg(api, job_name, num):
    return "Killing build: " + repr(job_name) + " - " + console_url(api, job_name, num)


def build_queued_msg(api, job_name, num):
    if api.api_type == ApiType.MOCK:
        queued_why = r"Why am I queued\?"
    else:
        queued_why = r"Build #[0-9]+ is already in progress \(ETA:([0-9.]+ sec|N/A)\)"
    return re.compile("^job: '" + job_name + "' Status QUEUED - " + queued_why)


_http_re = re.compile(r'https?://.*?/job/([^/" ]*)(/?)')
def replace_host_port(api_type, contains_url):
    if api_type != ApiType.SCRIPT:
        return _http_re.sub(r'http://x.x/job/\1\2', contains_url)
    else:
        return _http_re.sub(r'/tmp/jenkinsflow-test/job/\1.py', contains_url)


def lines_in(api_type, text, *expected_lines):
    """Assert that `*expected_lines` occur in order in the lines of `text`.

    Args:
        *expected_lines (str or RegexObject (hasattr `match`)): For each `expected_line` in expected_lines:
            If `expected_line` has a match method it is called and must return True for a line in `text`.
            Otherwise, if the `expected_line` starts with '^', a line in `text` must start with `expected_line[1:]`
            Otherwise `expected line` must simply occur in a line in `text`
    """

    assert isinstance(api_type, test_cfg.ApiType)
    assert text
    assert expected_lines

    return generic_lines_in(text, partial(replace_host_port, api_type), *expected_lines)


def flow_graph_dir(flow_name):
    """Control which directory to put flow graph in.

    Put the generated graph in the workspace root if running from Jenkins.
    If running from commandline put it under config.flow_graph_root_dir/flow_name

    Return: dir-name
    """
    return '.' if os.environ.get('JOB_NAME') else jp(flow_graph_root_dir, flow_name)


# Duplicated from set_build_result, we can't depend on that
jenkins_cli_jar = 'jenkins-cli.jar'
hudson_cli_jar = 'hudson-cli.jar'


def pre_existing_fake_cli(api_type):
    """Fake presense of cli-jar"""
    if api_type != ApiType.MOCK:
        return

    if os.environ.get('HUDSON_URL'):
        cli_jar = hudson_cli_jar
    else:
        cli_jar = jenkins_cli_jar

    with open(cli_jar, 'w'):
        pass



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pytest import raises

from jenkinsflow.flow import parallel, serial
from .framework import api_select


def test_flow_setup_exception_job(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(Exception) as exinfo:
            with serial(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                raise Exception("Not good")
            ctrl1.invoke('j1')

        assert str(exinfo.value) == "Not good"

        with raises(Exception) as exinfo:
            with parallel(api, 10, job_name_prefix=api.job_name_prefix) as ctrl1:
                raise Exception("Not good")
            ctrl1.invoke('j1')

        assert str(exinfo.value) == "Not good"



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import parallel, serial
from .framework import api_select


def test_single_level_serial(api_type):
    with api_select.api(__file__, api_type) as api:
        api.job('quick', max_fails=0, expect_invocations=1, expect_order=1, params=(('s1', 'Hi', 'desc'), ('c1', ('true', 'maybe', 'false'), 'desc')))
        api.job('wait10', max_fails=0, expect_invocations=1, expect_order=2, exec_time=10, serial=True)
        api.job('wait5', max_fails=0, expect_invocations=1, expect_order=3, exec_time=5, serial=True)

        with serial(api, timeout=40, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl:
            ctrl.invoke('quick', password='X', s1='HELLO', c1=True)
            ctrl.invoke('wait10')
            ctrl.invoke('wait5')


def test_single_level_parallel(api_type):
    with api_select.api(__file__, api_type) as api:
        api.job('quick', max_fails=0, expect_invocations=1, expect_order=1, params=(('s1', 'Hi', 'desc'), ('c1', ('true', 'maybe', 'false'), 'desc')))
        api.job('wait10', max_fails=0, expect_invocations=1, expect_order=1, exec_time=10)
        api.job('wait5', max_fails=0, expect_invocations=1, expect_order=1, exec_time=5)

        with parallel(api, timeout=20, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl:
            ctrl.invoke('quick', password='Y', s1='WORLD', c1='maybe')
            ctrl.invoke('wait10')
            ctrl.invoke('wait5')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

class UnBuffered(object):
    """
    Implements unbuffered output.
    """
    def __init__ (self, stream):
        self.stream = stream

    def write(self, data):
        self.stream.write(data)
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import re

import pytest
from pytest import raises

from jenkinsflow.flow import parallel, serial, MissingJobsException, FailedChildJobsException, FailedChildJobException, UnknownJobException
from .framework import api_select
from .framework.utils import lines_in
from .cfg import ApiType


def test_missing_jobs_not_allowed(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=0, expect_order=None)
        api.job('j2', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(MissingJobsException) as exinfo:
            with serial(api, 20, job_name_prefix=api.job_name_prefix) as ctrl1:
                ctrl1.invoke('j1')
                ctrl1.invoke('missingA')
                ctrl1.invoke('j2')

        assert lines_in(
            api_type, str(exinfo.value),
            re.compile("^Job not found: .*jenkinsflow_test__missing_jobs_not_allowed__missingA")
        )

        with raises(MissingJobsException):
            with serial(api, 20, job_name_prefix=api.job_name_prefix) as ctrl1:
                ctrl1.invoke('j1')
                ctrl1.invoke('missingA')
                with ctrl1.parallel() as ctrl2:
                    ctrl2.invoke('missingB')
                    ctrl2.invoke('j2')
                ctrl1.invoke('missingC')


def test_missing_jobs_allowed_still_missing_parallel(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=1)

        with raises(FailedChildJobsException):
            with parallel(api, 20, job_name_prefix=api.job_name_prefix, allow_missing_jobs=True) as ctrl1:
                ctrl1.invoke('j1')
                ctrl1.invoke('missingA')
                ctrl1.invoke('j2')


def test_missing_jobs_allowed_still_missing_serial(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j2', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(FailedChildJobException):
            with serial(api, 20, job_name_prefix=api.job_name_prefix, allow_missing_jobs=True) as ctrl1:
                ctrl1.invoke('j1')
                ctrl1.invoke('missingA')
                ctrl1.invoke('j2')


def test_missing_jobs_allowed_still_missing_parallel_serial(api_type):
    with api_select.api(__file__, api_type) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('j2', max_fails=0, expect_invocations=0, expect_order=None)

        with raises(FailedChildJobsException):
            with parallel(api, 20, job_name_prefix=api.job_name_prefix, allow_missing_jobs=True) as ctrl1:
                ctrl1.invoke('j1')
                ctrl1.invoke('missingA')
                with ctrl1.serial() as ctrl2:
                    ctrl2.invoke('missingB')
                    ctrl2.invoke('j2')
                ctrl1.invoke('missingC')


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: Handle ApiType.SCRIPT
def test_missing_jobs_allowed_created_serial_parallel(api_type):
    with api_select.api(__file__, api_type) as api:
        with api.job_creator():
            api.flow_job()
            api.job('j1', max_fails=0, expect_invocations=1, expect_order=1, create_job='missingA')
            api.job('missingA', max_fails=0, expect_invocations=1, expect_order=2, flow_created=True, create_job='missingB')
            api.job('missingB', max_fails=0, expect_invocations=1, expect_order=3, flow_created=True)
            api.job('j2', max_fails=0, expect_invocations=1, expect_order=3, create_job='missingC')
            api.job('missingC', max_fails=0, expect_invocations=1, expect_order=4, flow_created=True)

        with serial(api, 20, job_name_prefix=api.job_name_prefix, allow_missing_jobs=True) as ctrl1:
            ctrl1.invoke('j1')
            ctrl1.invoke('missingA')
            with ctrl1.parallel() as ctrl2:
                ctrl2.invoke('missingB')
                ctrl2.invoke('j2')
            ctrl1.invoke('missingC')

        # TODO: Validate output


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)  # TODO: Handle ApiTypes
def test_missing_jobs_job_disappeared(api_type):
    with api_select.api(__file__, api_type) as api:
        with api.job_creator():
            api.flow_job()
            api.job('j1', max_fails=0, expect_invocations=1, expect_order=1)
            api.job('disappearing', max_fails=0, expect_invocations=0, expect_order=None, disappearing=True)

        with raises(UnknownJobException):
            with serial(api, 20, job_name_prefix=api.job_name_prefix, allow_missing_jobs=True) as ctrl1:
                ctrl1.invoke('j1')
                ctrl1.invoke('disappearing')

        # TODO: Validate output



# Copyright (c) 2012 - 2017 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function

from .utils import base_url_and_api
from ..jenkins_cli_protocol import CliProtocol


def set_build_result(result, username, password, direct_url=None, java='java', protocol=CliProtocol.remoting):
    """Change the result of a Jenkins job.

    DEPRECATED - You should use the shell step exit code to determine the job result.

    Note: set_build_result can only be done from within the job, not after the job has finished.
    Note: Only available if URL is set in `Jenkins <http://jenkins-ci.org/>`_ system configuration.

    This command uses the Jenkins `cli` to change the result. It requires a java executable to run the Jenkins `cli`.
    Please note that in some versions of jenkins the cli is broken, it has no manifest file!

    Args:
        result (str): The result to set. Should probably be 'unstable'.
        username (str): Name of jenkins user with access to the job.
        password(str): Password of jenkins user with access to the job.
        direct-url (str): Jenkins URL. Default is JENKINS_URL/HUDSON_URL env var value. Use this argument if JENKINS_URL is a proxy.
        java (str): Alternative 'java' executable.
        protocol (CliProtocol): See https://jenkins.io/doc/book/managing/cli/.
    """

    base_url, api = base_url_and_api(direct_url)
    jenkins = api.Jenkins(base_url, username=username, password=password)
    jenkins.set_build_result(result, java, cli_call=True, protocol=protocol)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os, random

from jenkinsflow import jobload
from .framework import api_select


here = os.path.abspath(os.path.dirname(__file__))

_context = dict(
    exec_time=1,
    params=(),
    script=None,
    securitytoken='abc',
    print_env=False,
    create_job=None,
    num_builds_to_keep=4,
    final_result_use_cli=False,
    set_build_descriptions=()
)


def _random_job_name(api, short_name=None):
    # If short_name is not specified, use a random name to make sure the job doesn't exist
    short_name = short_name or str(random.random()).replace('.', '')
    return api.job_name_prefix + short_name, short_name


def _assert_job(api, job_name, cleanup=False):
    job = api.get_job(job_name)
    assert job is not None
    assert job.name == job_name
    assert job.public_uri is not None and job_name in job.public_uri

    if cleanup:
        api.delete_job(job_name)
        return None

    return job


def test_job_load_new_no_pre_delete(api_type):
    api = api_select.api(__file__, api_type, login=True)
    full_name, short_name = _random_job_name(api)
    api.job(short_name, 1, 1, 1, exec_time=1, non_existing=True)
    jobload.update_job_from_template(api, full_name, api.job_xml_template, pre_delete=False, context=_context)
    _assert_job(api, full_name, cleanup=True)


def test_job_load_new_pre_delete(api_type):
    api = api_select.api(__file__, api_type, login=True)
    full_name, short_name = _random_job_name(api)
    api.job(short_name, 1, 1, 1, exec_time=1, non_existing=True)
    jobload.update_job_from_template(api, full_name, api.job_xml_template, pre_delete=True, context=_context)
    _assert_job(api, full_name, cleanup=True)


def test_job_load_existing_pre_delete(api_type):
    api = api_select.api(__file__, api_type, login=True)
    full_name, short_name = _random_job_name(api)
    api.job(short_name, 1, 1, 1, exec_time=1)
    jobload.update_job_from_template(api, full_name, api.job_xml_template, pre_delete=True, context=_context)
    _assert_job(api, full_name, cleanup=False)
    jobload.update_job_from_template(api, full_name, api.job_xml_template, pre_delete=True, context=_context)
    _assert_job(api, full_name, cleanup=True)


def test_job_load__existing_update(api_type):
    api = api_select.api(__file__, api_type, login=True)
    full_name, short_name = _random_job_name(api)
    api.job(short_name, 1, 1, 1, exec_time=1)
    jobload.update_job_from_template(api, full_name, api.job_xml_template, pre_delete=True, context=_context)
    _assert_job(api, full_name, cleanup=False)
    jobload.update_job_from_template(api, full_name, api.job_xml_template, pre_delete=False, context=_context)
    _assert_job(api, full_name, cleanup=True)


def test_job_load_non_existing_pre_delete(api_type):
    api = api_select.api(__file__, api_type, login=True)
    full_name, short_name = _random_job_name(api)
    api.job(short_name, 1, 1, 1, exec_time=1, non_existing=True)
    jobload.update_job_from_template(api, full_name, api.job_xml_template, pre_delete=True, context=_context)



# -*- coding: utf-8 -*-

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function

import sys, os, codecs
major_version = sys.version_info.major
from os.path import join as jp

from jenkinsflow.flow import serial

from .framework import api_select
from .cfg import ApiType


_here = os.path.dirname(os.path.abspath(__file__))


from .set_build_description_test import _clear_description


def _verify_description(api, job, build_number, expected):
    if api.api_type == ApiType.MOCK:
        return

    # Read back description and verify
    if api.api_type == ApiType.JENKINS:
        build_url = "/job/" + job.name + '/' + str(build_number)
        dct = api.get_json(build_url, tree="description")
        description = dct['description']

    if api.api_type == ApiType.SCRIPT:
        with codecs.open(jp(job.workspace, 'description.txt'), encoding='utf-8') as ff:
            description = ff.read()

    assert description == expected


def test_set_build_description_unicode_set_build_description_util(api_type):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        job_name = 'job-1'
        api.job(job_name, max_fails=0, expect_invocations=1, expect_order=1)

        # Need to read the build number
        if api.api_type == ApiType.SCRIPT:
            # TODO: This can't be called here for Jenkins API. Why?
            job = api.get_job(api.job_name_prefix + job_name)
            _clear_description(api, job)

        with serial(api, timeout=70, job_name_prefix=api.job_name_prefix, report_interval=1, description="«©º") as ctrl1:
            ctrl1.invoke(job_name, password='a', s1='b')

        if api.api_type != ApiType.SCRIPT:
            job = api.get_job(api.job_name_prefix + job_name)
        _, _, build_num = job.job_status()

        api.set_build_description(u'ÆØÅ', job_name=job.name, build_number=build_num)
        api.set_build_description(u'æøå', replace=False, job_name=job.name, build_number=build_num)
        _verify_description(api, job, build_num, u'«©º\nÆØÅ\næøå')
        
        api.set_build_description(u'¶¹²', replace=True, job_name=job.name, build_number=build_num)
        api.set_build_description(u'³¼¢', replace=False, separator='#', job_name=job.name, build_number=build_num)
        api.set_build_description(u'¬<>©‘’Nº', separator='!!', job_name=job.name, build_number=build_num)
        _verify_description(api, job, build_num, u'¶¹²#³¼¢!!¬<>©‘’Nº')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function

import sys, os, signal, time
major_version = sys.version_info.major
if major_version < 3:
    import subprocess32 as subprocess
else:
    import subprocess

from jenkinsflow.test.framework.logger import log, logt


def _killer(log_file, pid, sleep_time, num_kills):
    log(log_file, '\n')
    logt(log_file, "Killer going to sleep for", sleep_time, "seconds")
    time.sleep(sleep_time)
    logt(log_file, "Killer woke up")
    for ii in range(0, num_kills):
        logt(log_file, "Killer sending", ii + 1, "of", num_kills, "SIGTERM signals to ", pid)
        os.kill(pid, signal.SIGTERM)
        logt(log_file, "Killer sent", ii + 1, "of", num_kills, "SIGTERM signals to ", pid)
        time.sleep(1)


if __name__ == '__main__':
    log_file_name = sys.argv[4]
    with open(log_file_name, 'a+') as log_file:
        _killer(log_file, int(sys.argv[1]), float(sys.argv[2]), int(sys.argv[3]))


def kill(api, sleep_time, num_kills):
    """Kill this process"""
    pid = os.getpid()    
    ff = __file__.replace('.pyc', '.py')
    log_file_name = api.func_name.replace('test_', '')
    args = [sys.executable, ff, repr(pid), repr(sleep_time), repr(num_kills), log_file_name]
    with open(log_file_name, 'w') as log_file:
        logt(log_file, "Invoking kill subprocess.", args)

    subprocess.Popen(args)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os, re

import pytest
from pytest import raises

from jenkinsflow.flow import parallel, FailedChildJobsException

from .framework import api_select
from .framework.utils import lines_in
from .framework.abort_job import abort

from .cfg import ApiType

here = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.not_apis(ApiType.SCRIPT)
def test_abort(api_type, capsys):
    with api_select.api(__file__, api_type, login=True) as api:
        api.flow_job()
        api.job('quick', max_fails=0, expect_invocations=1, expect_order=1)
        api.job('wait10_abort', max_fails=0, expect_invocations=1, expect_order=1, exec_time=20, final_result='ABORTED')
        api.job('wait1_fail', max_fails=1, expect_invocations=1, expect_order=1, exec_time=1)

        abort(api, 'wait10_abort', 10)

        with raises(FailedChildJobsException) as exinfo:
            with parallel(api, timeout=40, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl:
                ctrl.invoke('quick')
                ctrl.invoke('wait10_abort')
                ctrl.invoke('wait1_fail')

        assert "wait10_abort" in str(exinfo.value)
        assert "wait1_fail" in str(exinfo.value)

        sout, _ = capsys.readouterr()
        assert lines_in(
            api_type, sout,
            re.compile("^ABORTED: 'jenkinsflow_test__abort__wait10_abort' - build: .*/jenkinsflow_test__abort__wait10_abort.* after:"),
        )



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import re
import pytest

from jenkinsflow.flow import parallel, serial, FailedChildJobException, FailedChildJobsException
from .framework import api_select
from .framework.utils import lines_in
from .cfg import ApiType


@pytest.mark.not_apis(ApiType.MOCK)
def test_replace_invocation_class_log_override(api_type, capsys):
    if api_type == ApiType.JENKINS:
        from jenkinsflow.jenkins_api import Invocation

        class LogInvocation(Invocation):
            def console_url(self):
                return "HELLO LOG"

    elif api_type == ApiType.SCRIPT:
        from jenkinsflow.script_api import Invocation

        class LogInvocation(Invocation):
            def console_url(self):
                return "HELLO LOG"
    else:
        raise Exception("Invalid ApiType: " + repr(api_type))

    with api_select.api(__file__, api_type, invocation_class=LogInvocation) as api:
        api.flow_job()
        api.job('j1', max_fails=0, expect_invocations=1, expect_order=1, params=(('s1', '', 'desc'), ('c1', 'false', 'desc')))
        api.job('j2', max_fails=0, expect_invocations=1, expect_order=1, params=(('s1', '', 'desc'), ('c1', 'false', 'desc')))

        with parallel(api, timeout=40, job_name_prefix=api.job_name_prefix, report_interval=3) as ctrl:
            ctrl.invoke('j1', s1='', c1=False)
            ctrl.invoke('j2', s1='', c1=False)

        sout, _ = capsys.readouterr()
        assert "HELLO LOG" in sout





# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os

import pytest
from pytest import fixture  # pylint: disable=no-name-in-module
from click.testing import CliRunner

from . import cfg as test_cfg
from .cfg import ApiType

# Note: You can't (indirectly) import stuff from jenkinsflow here, it messes up the coverage


@pytest.fixture(params=[ApiType.MOCK, ApiType.JENKINS, ApiType.SCRIPT])
def api_type(request):
    """ApiType fixture"""
    selected_api_type = request.param    

    apimarker = request.node.get_marker("apis")
    if apimarker is not None:
        apis = apimarker.args
        if selected_api_type not in apis:
            pytest.skip("test requires one the following apis {apis}, current {api_type}".format(apis=apis, api_type=selected_api_type))

    not_apimarker = request.node.get_marker("not_apis")
    if not_apimarker is not None:
        not_apis = not_apimarker.args
        if selected_api_type in not_apis:
            pytest.skip("test is not run for the following apis {apis}, current {api_type}".format(apis=not_apis, api_type=selected_api_type))

    return selected_api_type


def pytest_configure(config):
    # Register api  marker
    config.addinivalue_line("markers", "apis(*ApiType): mark test to run only when using specified apis")
    config.addinivalue_line("markers", "not_apis(*ApiType): mark test NOT to run when using specified apis")


def _set_env_fixture(var_name, value, request):
    """
    Ensure env var_name is set to the value <value>
    Set back to original value, if any, or unset it, after test.
    """
    has_var = os.environ.get(var_name)
    os.environ[var_name] = value
    if not has_var:
        def fin():
            del os.environ[var_name]
    else:
        def fin():
            os.environ[var_name] = has_var
    request.addfinalizer(fin)


def _set_jenkins_url_env_fixture(not_set_value, request):
    if os.environ.get('JENKINS_URL'):
        _set_env_fixture('JENKINS_URL', not_set_value, request)
        return

    _set_env_fixture('HUDSON_URL', not_set_value, request)


def _set_env_if_not_set_fixture(var_name, not_set_value, request):
    """
    Ensure env var_name is set to the value 'not_set_value' IFF it was not already set.
    Unset it after test.
    """
    has_var = os.environ.get(var_name)
    if not has_var:
        os.environ[var_name] = not_set_value
        def fin():
            del os.environ[var_name]
        request.addfinalizer(fin)


def _set_jenkins_url_env_if_not_set_fixture(not_set_value, request):
    if not os.environ.get('HUDSON_URL'):
        _set_env_if_not_set_fixture('JENKINS_URL', not_set_value, request)


def _unset_env_fixture(var_name, request):
    """
    Ensure env var_name is NOT set
    Set back to original value, if any, after test.
    """
    has_var = os.environ.get(var_name)
    if has_var:
        del os.environ[var_name]
        def fin():
            os.environ[var_name] = has_var
        request.addfinalizer(fin)


@fixture
def env_base_url(request, api_type):
    # Fake that we are running from inside jenkins job
    public_url = test_cfg.public_url(api_type)
    _set_jenkins_url_env_if_not_set_fixture(public_url, request)
    return public_url


@fixture
def env_base_url_trailing_slash(request, api_type):
    _set_jenkins_url_env_if_not_set_fixture(test_cfg.public_url(api_type) + '/', request)


@fixture
def env_base_url_trailing_slashes(request, api_type):
    _set_jenkins_url_env_if_not_set_fixture(test_cfg.public_url(api_type) + '//', request)


@fixture
def env_no_base_url(request):
    # Make sure it looks as if we are we are running from outside jenkins job
    _unset_env_fixture('JENKINS_URL', request)
    _unset_env_fixture('HUDSON_URL', request)


@fixture
def env_different_base_url(request):
    # Fake that we are running from inside jenkins job
    # This url is not used, but should simply be different fron direct_url used in test, to simulate proxied jenkins
    _set_jenkins_url_env_fixture(test_cfg.proxied_public_url, request)


@fixture
def env_job_name(request):
    # Fake that we are running from inside jenkins job
    _set_env_if_not_set_fixture('JOB_NAME', 'hudelihuu', request)


@fixture
def env_build_number(request):
    # Fake that we are running from inside jenkins job
    _set_env_if_not_set_fixture('BUILD_NUMBER', '1', request)


@fixture(scope="module")
def fake_java(request):
    if not os.environ.get('BUILD_URL'):
        # Running outside of Jenkins, fake call to java - cli, use script ./framework/java
        here = os.path.abspath(os.path.dirname(__file__))
        orig_path = os.environ.get('PATH')
        os.environ['PATH'] = os.path.join(here, 'framework') + ':' + orig_path or ''

        if orig_path:
            def fin():
                os.environ['PATH'] = orig_path
            request.addfinalizer(fin)


@fixture(scope='function')
def cli_runner(request):
    return CliRunner()



from __future__ import print_function

import sys, os

from setuptools import setup
from setuptools.command.test import test as TestCommand


PROJECT_ROOT, _ = os.path.split(__file__)
PROJECT_NAME = 'jenkinsflow'
COPYRIGHT = u"Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT"
PROJECT_AUTHORS = u"Lars Hupfeldt Nielsen"
PROJECT_EMAILS = 'lhn@hupfeldtit.dk'
PROJECT_URL = "https://github.com/lhupfeldt/jenkinsflow"
SHORT_DESCRIPTION = 'Python API with high level build flow constructs (parallel/serial) for Jenkins (and Hudson).'
LONG_DESCRIPTION = open(os.path.join(PROJECT_ROOT, "README.rst")).read()

on_rtd = os.environ.get('READTHEDOCS') == 'True'
is_ci = os.environ.get('CI', 'false').lower() == 'true'


class Test(TestCommand):
    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import test.run
        if is_ci:
            print("Running under CI")
            # Note 'mock' is also hardcoded in .travis.yml
            sys.exit(test.run.cli(apis='mock,script', mock_speedup=10))
        sys.exit(test.run.cli(apis='mock,script'))


flow_requires = ['atomicfile>=1.0,<=2.0']
cli_requires = ['click>=6.0']
job_load_requires = ['tenjin>=1.1.1']
jenkins_api_requires = ['requests>=2.7,<=3.0']
# You need to install python(3)-devel to be be able to install psutil, see INSTALL.md
script_api_requires = ['psutil>=5.2.2', 'setproctitle>=1.1.10']
visual_requires = ['bottle>=0.12.1']

if sys.version_info.major < 3:
    flow_requires.extend(['enum34>=1.1', 'subprocess32>=3.2.5'])

if not on_rtd:
    install_requires = flow_requires + cli_requires + job_load_requires + jenkins_api_requires + script_api_requires + visual_requires
else:
    install_requires = flow_requires + cli_requires + jenkins_api_requires + script_api_requires

tests_require = [
    'pytest>=3.0.5', 'pytest-cov>=2.4.0', 'pytest-instafail>=0.3.0', 'pytest-xdist>=1.16',
    'click>=6.0', 'tenjin>=1.1.1', 'bottle>=0.12',
    # The test also tests creation of the documentation
    'sphinx>=1.6.1', 'sphinxcontrib-programoutput']

if sys.version_info.major < 3:
    tests_require.append('proxytypes')
else:
    tests_require.append('objproxies>=0.9.4')

extras = {
    'test': tests_require,
}

if __name__ == "__main__":
    setup(
        name=PROJECT_NAME.lower(),
        version_command=('git describe', 'pep440-git'),
        author=PROJECT_AUTHORS,
        author_email=PROJECT_EMAILS,
        packages=['jenkinsflow', 'jenkinsflow.utils', 'jenkinsflow.cli'],
        package_dir={'jenkinsflow':'.', 'jenkinsflow.utils': 'utils', 'jenkinsflow.cli': 'cli'},
        zip_safe=True,
        include_package_data=False,
        install_requires=install_requires,
        setup_requires='setuptools-version-command>=2.2',
        test_suite='test',
        tests_require=tests_require,
        extras_require=extras,
        cmdclass={'test': Test},
        url=PROJECT_URL,
        description=SHORT_DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        license='BSD',
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Natural Language :: English',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Topic :: Software Development :: Testing',
        ],
        entry_points='''
            [console_scripts]
            jenkinsflow=jenkinsflow.cli.cli:cli
        ''',
    )



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function

import abc
from collections import OrderedDict
import os
from os.path import join as jp

from jenkinsflow.api_base import UnknownJobException, BuildResult, Progress
from .abstract_api import AbstractApiJob, AbstractApiJenkins
from .config import test_tmp_dir


def _mkdir(path):
    try:
        os.mkdir(path)
    except OSError:
        if not os.path.exists(path):
            raise


class TestJob(AbstractApiJob):
    __metaclass__ = abc.ABCMeta

    _current_order = 1

    def __init__(self, exec_time, max_fails, expect_invocations, expect_order, initial_buildno, invocation_delay,
                 unknown_result, final_result, serial, print_env, flow_created, create_job, disappearing, non_existing, kill):
        """
        Set unknown_result to True if the result is indeterminate (timeout or invoke_unchecked)
        """
        assert exec_time > 0
        assert max_fails >= 0
        assert expect_invocations >= 0
        assert expect_order is None or expect_order >= 1
        assert initial_buildno is None or initial_buildno >= 1
        assert invocation_delay >= 0
        assert unknown_result in (False, True)
        assert serial in (False, True)
        assert flow_created in (False, True)
        assert disappearing in (False, True)
        assert non_existing in (False, True)

        self.exec_time = exec_time
        self.max_fails = max_fails
        self.expect_invocations = expect_invocations
        self.expect_order = expect_order
        self.initial_buildno = initial_buildno
        self.invocation_delay = invocation_delay
        self.unknown_result = unknown_result
        self.serial = serial
        self.print_env = print_env
        self.final_result = final_result if isinstance(final_result, (BuildResult, type(None))) else BuildResult[final_result.upper()]
        self.flow_created = flow_created
        self.create_job = create_job
        self.disappearing = disappearing
        self.non_existing = non_existing
        self.kill = kill

        self.invocation_number = 0
        self.invocation_time = self.start_time = self.end_time = 0
        self.actual_order = -1

        self.build_params = None

    @property
    def has_force_result_param(self):
        return self.max_fails > 0 or self.final_result

    def invoke(self, securitytoken, build_params, cause, description):
        self.build_params = build_params
        self.invocation_number += 1
        self.actual_order = TestJob._current_order
        TestJob._current_order += 1

    def __repr__(self):
        return ", expect_invocations: " + repr(self.expect_invocations) + \
            ", invocation_number: " + repr(self.invocation_number) + \
            ", expect_order: " + repr(self.expect_order) + \
            ", start_time: " + repr(self.start_time) + \
            ", exec_time: " + repr(self.exec_time) + \
            ", end_time: " + repr(self.end_time)


class Jobs(object):
    def __init__(self, api):
        self.api = api

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            return None

        test_jobs = self.api.test_jobs
        for job_name, job in test_jobs.items():
            if job.flow_created:
                # Patch up another job that is supposed to be created by this job - replace job_name with job reference
                for other_job in test_jobs.values():
                    if isinstance(other_job.create_job, str):
                        other_job_name = self.api.job_name_prefix + other_job.create_job
                        if other_job_name == job_name:
                            other_job.create_job = job
                            break
                else:
                    raise Exception("Job: " + repr(job_name) + " is supposed to be created by another job, but that job was not found")

        for job_name, job in test_jobs.items():
            if job.create_job and isinstance(job.create_job, str):
                raise Exception("Job: " + repr(job_name) + " is supposed to create job: " + repr(job.create_job) + " but definition for that job was not found")


class TestJenkins(AbstractApiJenkins):
    __metaclass__ = abc.ABCMeta

    def __init__(self, job_name_prefix, **kwargs):
        super(TestJenkins, self).__init__(**kwargs)
        self.job_name_prefix = job_name_prefix
        TestJob._current_order = 1
        self.test_jobs = OrderedDict()

    @abc.abstractmethod
    def job(self, name, max_fails, expect_invocations, expect_order, exec_time=None, initial_buildno=None, invocation_delay=0.1, params=None,
            script=None, unknown_result=False, final_result=None, serial=False, print_env=False, flow_created=False, create_job=None, disappearing=False,
            non_existing=False, kill=False, num_builds_to_keep=4, allow_running=False):
        """Create a job with corresponding test metadata.

        Args:
            name (str): Base name of job, the test framework will add a prefix base on test module
            exec_time (int): Number of seconds that the job will run (sleep), actual run may/will be longer
            max_fails (int): Number of times the job will fail during this flow (when using retry)
            expect_invocations (int): Number of expected invocation during this flow. Will be larger or equal to exec_time.
            ...
            flow_created (boolean): This job is expected to be non-existing at start of flow and be created during the flow
            create_job (str): Name of another job that will be created by this job, when this job is running
        """
        pass

    # Delete/Create hack sufficient to get resonable coverage on job_load test
    def delete_job(self, job_name):
        try:
            job = self.test_jobs[job_name]
            if job.non_existing:
                raise UnknownJobException(job_name)
            job.non_existing = True
        except KeyError:
            raise Exception("Test job setup error, missing test job definition:", job_name)

    def create_job(self, job_name, config_xml):
        try:
            job = self.test_jobs[job_name]
            job.non_existing = False
        except KeyError:
            raise Exception("Test job setup error, missing test job definition:", job_name)

    def job_creator(self):
        return Jobs(self)

    @abc.abstractmethod
    def flow_job(self, name=None, params=None):
        pass

    def flow_job_name(self, name):
        # Don't create flow jobs when mocked
        name = '0flow_' + name if name else '0flow'
        return (self.job_name_prefix or '') + name

    def __enter__(self):
        # pylint: disable=attribute-defined-outside-init
        self._pre_work_dir = os.getcwd()
        self._work_dir = jp(test_tmp_dir,self.job_name_prefix)
        _mkdir(self._work_dir)
        os.chdir(self._work_dir)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(self._pre_work_dir)
        if not exc_type:
            self.test_results()

    def test_results(self):
        print("Jenkinsflow Test Framework -- Checking results")

        max_actual_order = 0
        last_expected_order = 0
        last_end_time = 0

        for job in self.test_jobs.values():
            if job.expect_order is not None:
                # Check job invocation order
                assert last_expected_order <= job.expect_order, "Mock job list must be sorted by expected order, error in test setup."

                assert job.actual_order >= job.expect_order, "Job: " + job.name + " was started out of order, " + \
                    "job.actual_order: " + repr(job.actual_order) + ", job.expect_order: " + repr(job.expect_order)

                if job.expect_order > last_expected_order:
                    assert job.actual_order > max_actual_order

                if job.serial:
                    assert job.invocation_time >= last_end_time, "Serial job " + job.name + " started before previous job finished"
                last_end_time = job.end_time

                last_expected_order = job.expect_order
                max_actual_order = max(job.actual_order, max_actual_order)

            if  job.expect_invocations is not None:
                # Check expected number of job invocations
                assert job.expect_invocations == job.invocation_number, "Job: " + job.name + " invoked " + str(job.invocation_number) + " times, expected " + str(job.expect_invocations) + " invocations"

            if job.unknown_result:
                # The job must still be running, but maybe it has not been started yet, so wait up to 3 seconds for it to start
                max_wait_to_start = 3
                start_check_interval = 0.01
                max_wait_index = int(max_wait_to_start / start_check_interval)
                prev_progress = None
                for ii in range(1, max_wait_index):
                    # TODO: job obj should be invocation obj!
                    result, progress, last_build_number = job.job_status()
                    if progress == Progress.RUNNING:
                        break
                    if progress != prev_progress:
                        print("FW: last build status:", job.name, result, progress, last_build_number)
                        prev_progress = progress
                    self.sleep(start_check_interval)
                    if hasattr(job, 'jenkins'):
                        job.jenkins.quick_poll()
                    job.poll()

                # pylint: disable=maybe-no-member
                assert progress == Progress.RUNNING, "Job: " + job.name + " is expected to be running, but state is " + progress.name
                # Now stop the job, so that it won't be running after the testsuite is finished
                job.stop_all()
            elif job.expect_invocations != 0 and not job.kill:
                if job.invocation_number > job.max_fails:
                    expect_status = BuildResult.SUCCESS if job.final_result is None else job.final_result
                else:
                    expect_status = BuildResult.FAILURE
                # TODO job obj should be called invocation!
                # Get last obj
                try:
                    key = next(reversed(job._invocations))
                    invocation = job._invocations[key]
                except TypeError:
                    # script_api ?
                    invocation = job._invocations[-1]

                assert invocation.build_number is not None, "Job: " + repr(job) + " should have had build_number, but it has None"
                result, progress = invocation.status()
                assert result == expect_status, "Job: " + job.name + " expected result " + repr(expect_status) + " but got " + repr(result)
            elif job.kill:
                result, progress, _ = job.job_status()
                assert progress == Progress.IDLE
                # TODO
                # expect_status = BuildResult.ABORTED
                # assert result == expect_status, "Job: " + job.name + " expected result " + repr(expect_status) + " but got " + repr(result)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from __future__ import print_function

import pytest

from .. import jenkins_api
from .framework import api_select
from .cfg import ApiType


@pytest.mark.not_apis(ApiType.MOCK, ApiType.SCRIPT)
def test_api_class_repr_job(api_type):
    api = api_select.api(__file__, api_type, login=True)
    job = jenkins_api.ApiJob(api, {}, 'my-job')

    jrd = eval(repr(job))
    assert jrd == {'name': 'my-job', 'dct': {}}

    invocation = jenkins_api.Invocation(job, "http://dummy", 'hello')
    assert repr(invocation) == "Invocation: 'http://dummy' None None"
    



import time as real_time


class HyperSpeed(object):
    def __init__(self, speedup, **kwargs):
        super(HyperSpeed, self).__init__()
        assert isinstance(speedup, int)
        self.speedup = speedup

    def time(self):
        return real_time.time() * self.speedup

    def sleep(self, seconds):
        return real_time.sleep(seconds / float(self.speedup))



# Copyright (c) 2012 - 2017 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.


from enum import Enum


class CliProtocol(Enum):
    http = 0
    ssh = 1
    remoting = 2
